/*
 * Name:    d_aps_10id.c 
 *
 * Purpose: MX motor driver to control the MRCAT Sector 10-ID monochromator
 *          at the Advanced Photon Source.  This is a pseudomotor that 
 *          pretends to be the monochromator 'theta' axis.  The value
 *          written to this pseudo motor is used to control the real 'theta'
 *          axis and any other MX devices that are tied into this pseudomotor.
 *
 * Note:    This driver has been replaced by the 'monochromator' pseudomotor
 *          driver, but is still included as a pseudomotor example.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_variable.h"
#include "mx_alloc.h"
#include "mx_constants.h"
#include "mx_motor.h"
#include "d_aps_10id.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_aps_10id_motor_record_function_list = {
	mxd_aps_10id_motor_initialize_type,
	mxd_aps_10id_motor_create_record_structures,
	mxd_aps_10id_motor_finish_record_initialization,
	mxd_aps_10id_motor_delete_record,
	mxd_aps_10id_motor_print_motor_structure,
	mxd_aps_10id_motor_read_parms_from_hardware,
	mxd_aps_10id_motor_write_parms_to_hardware,
	mxd_aps_10id_motor_open,
	mxd_aps_10id_motor_close
};

MX_MOTOR_FUNCTION_LIST mxd_aps_10id_motor_motor_function_list = {
	mxd_aps_10id_motor_motor_is_busy,
	mxd_aps_10id_motor_move_absolute,
	mxd_aps_10id_motor_get_position,
	mxd_aps_10id_motor_set_position,
	mxd_aps_10id_motor_soft_abort,
	mxd_aps_10id_motor_immediate_abort,
	mxd_aps_10id_motor_positive_limit_hit,
	mxd_aps_10id_motor_negative_limit_hit,
	mxd_aps_10id_motor_find_home_position
};

/* APS_10ID motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_aps_10id_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_APS_10ID_MOTOR_STANDARD_FIELDS
};

long mxd_aps_10id_motor_num_record_fields
		= sizeof( mxd_aps_10id_motor_record_field_defaults )
			/ sizeof( mxd_aps_10id_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_aps_10id_motor_rfield_def_ptr
			= &mxd_aps_10id_motor_record_field_defaults[0];

/* === */

MX_EXPORT mx_status_type
mxd_aps_10id_motor_initialize_type( long type )
{
		/* Nothing needed here. */

		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_aps_10id_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_APS_10ID_MOTOR *aps_10id_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_APS_10ID_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = aps_10id_motor;
	record->class_specific_function_list
				= &mxd_aps_10id_motor_motor_function_list;

	motor->record = record;

	/* A DAC motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	return status;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_print_motor_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_aps_10id_motor_print_motor_structure()";

	MX_MOTOR *motor;
	MX_APS_10ID_MOTOR *aps_10id_motor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *) (record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type                 = APS_10ID_MOTOR.\n\n");

	fprintf(file, "  name                       = %s\n", record->name);
	fprintf(file, "  real theta motor           = %s\n",
				aps_10id_motor->real_theta_motor_record->name);
	fprintf(file, "  omega motor                = %s\n",
				aps_10id_motor->omega_motor_record->name);
	fprintf(file, "  omega function type record = %s\n",
			aps_10id_motor->omega_function_type_record->name);
	fprintf(file, "  omega parameters record    = %s\n",
				aps_10id_motor->omega_parameters_record->name);
	fprintf(file, "  gap energy motor           = %s\n",
				aps_10id_motor->gap_energy_motor_record->name);
	fprintf(file, "  gap linked                 = %s\n",
				aps_10id_motor->gap_linked_record->name);
	fprintf(file, "  gap offset                 = %s\n",
				aps_10id_motor->gap_offset_record->name);
	fprintf(file, "  gap harmonic               = %s\n",
				aps_10id_motor->gap_harmonic_record->name);
	fprintf(file, "  d spacing                  = %s\n",
				aps_10id_motor->d_spacing_record->name);
	fprintf(file, "  angle scale                = %g\n",
				aps_10id_motor->angle_scale);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position                   = %g %s  (%g).\n",
		motor->position, motor->units,
		motor->raw_position.analog );
	fprintf(file, "  theta scale                = %g %s "
					"per unscaled theta unit.\n",
		motor->scale, motor->units);
	fprintf(file, "  theta offset               = %g %s.\n",
		motor->offset, motor->units);
	fprintf(file, "  backlash                   = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );
	fprintf(file, "  negative limit             = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );
	fprintf(file, "  positive limit             = %g %s  (%g).\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.analog );

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband                   = %g %s  (%g).\n",
		deadband, motor->units,
		motor->deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_read_parms_from_hardware( MX_RECORD *record )
{
	/* All saving of parameters is handled by the MX_MOTOR record
	 * this motor is derived from, so we need not do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_write_parms_to_hardware( MX_RECORD *record )
{
	/* All setting of parameters is handled by the MX_MOTOR record
	 * this motor is derived from, so we need not do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_open( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_close( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_motor_is_busy()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *real_theta_motor_record;
	MX_RECORD *omega_motor_record;
	MX_RECORD *gap_energy_motor_record;
	int busy;
	long gap_linked;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	/* Check to see if theta is busy. */

	real_theta_motor_record = aps_10id_motor->real_theta_motor_record;

	if ( real_theta_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"real_theta_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_is_busy( real_theta_motor_record, &busy );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( busy ) {
		motor->busy = TRUE;

		return MX_SUCCESSFUL_RESULT;
	}

	/* If theta was not busy, check to see if omega is busy. */

	omega_motor_record = aps_10id_motor->omega_motor_record;

	if ( omega_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"omega_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_is_busy( omega_motor_record, &busy );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( busy ) {
		motor->busy = TRUE;

		return MX_SUCCESSFUL_RESULT;
	}

	/* If theta and omega were not busy, check to see if the gap motor
	 * is busy.
	 */

	/* First check to see whether or not the gap is supposed to be
	 * linked to theta motions.
	 */

	status = mx_get_long_variable_by_name(
				 real_theta_motor_record->list_head,
				aps_10id_motor->gap_linked_record->name,
				&gap_linked );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( gap_linked != 0 ) {
		gap_energy_motor_record =
			aps_10id_motor->gap_energy_motor_record;

		if ( gap_energy_motor_record == (MX_RECORD *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"gap_energy_motor_record pointer for record '%s' is NULL.",
				motor->record->name );
		}

		status = mx_motor_is_busy( gap_energy_motor_record, &busy );
	}

	motor->busy = busy;

	return status;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_move_absolute()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *list_head;
	int num_motors;
	MX_RECORD *motor_record_array[MXF_APS_10ID_NUM_REAL_MOTOR_RECORDS];
	double position_array[MXF_APS_10ID_NUM_REAL_MOTOR_RECORDS];
	MX_RECORD *real_theta_motor_record;
	MX_RECORD *omega_motor_record;
	double omega_old_position;
	int real_theta_motor_index, omega_motor_index, gap_energy_motor_index;
	MX_RECORD_FIELD *value_field;
	void *data_ptr;
	double *parameter_array;
	long num_parameters;
	long function_type;
	double theta, omega;
	double angle_in_radians;
	long gap_linked;
	double d_spacing, gap_offset, mono_energy, new_gap_energy;
	long gap_harmonic;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	list_head = motor->record->list_head;

	/* First check to see whether or not the gap is supposed to be
	 * linked to theta motions.  This determines the array index
	 * where values are to be written into motor_record_array
	 * and position_array.  We try to make the omega motor always
	 * be the last one to move, since it is the only one that may
	 * need to do a backlash correction ( and backlash corrections
	 * are currently handled synchronously ).
	 */

	status = mx_get_long_variable_by_name( list_head,
				aps_10id_motor->gap_linked_record->name,
				&gap_linked );

	if ( status.code != MXE_SUCCESS )
		return status;

	real_theta_motor_index = 0;

	if ( gap_linked == FALSE ) {
		num_motors = 2;

		gap_energy_motor_index = -1;
		omega_motor_index = 1;
	} else {
		num_motors = 3;

		gap_energy_motor_index = 1;
		omega_motor_index = 2;
	}

	/***/

	motor_record_array[ real_theta_motor_index ]
				= aps_10id_motor->real_theta_motor_record;

	if (motor_record_array[ real_theta_motor_index ] == (MX_RECORD *)NULL){
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Real theta motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	motor_record_array[ omega_motor_index ]
				= aps_10id_motor->omega_motor_record;

	if (motor_record_array[ omega_motor_index ] == (MX_RECORD *) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Omega motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	if ( gap_linked ) {
		motor_record_array[ gap_energy_motor_index ]
				= aps_10id_motor->gap_energy_motor_record;

		if (motor_record_array[ gap_energy_motor_index ]
						== (MX_RECORD *) NULL) {

			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Gap energy motor record pointer for record '%s' is NULL.",
				motor->record->name );
		}
	}

	/***************** Evaluate new position for theta ****************/

	/* This part is simple.  We just take the pseudomotor destination
	 * and copy it to the real motor destination.
	 */

	theta = motor->raw_destination.analog;

	position_array[ real_theta_motor_index ] = theta;

	/***************** Evaluate new position for omega ****************/

	/* Get the fitting function type. */

	status = mx_get_long_variable_by_name( list_head,
			aps_10id_motor->omega_function_type_record->name,
				&function_type );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the omega motion parameters.  They should be in a 
	 * one dimensional array of doubles.
	 */

	status = mx_find_record_field( aps_10id_motor->omega_parameters_record,
					"value", &value_field );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_parameters = value_field->dimension[0];

	data_ptr = value_field->data_pointer;

	if ( value_field->flags && MXFF_VARARGS ) {
		parameter_array = (double *)
			mx_read_void_pointer_from_memory_location( data_ptr );
	} else {
		parameter_array = (double *) data_ptr;
	}

	/* Get the current position of omega. */

        omega_motor_record = motor_record_array[ omega_motor_index ];

	status = mx_motor_get_position(
				omega_motor_record, &omega_old_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Evaluate the function using the values in 'parameter_array'. */

	switch( function_type ) {
	case MXF_APS_10ID_NO_DEPENDENCE:
		omega = omega_old_position;
		break;
	case MXF_APS_10ID_CUBIC_POLYNOMIAL_FIT:
		if ( num_parameters < 4 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"A cubic polynomial fit requires 4 parameters.  "
			"'%s' has only %ld parameters.",
				aps_10id_motor->omega_parameters_record->name,
				num_parameters );
		}
		omega = parameter_array[0] + theta * (
			parameter_array[1] + theta * (
			parameter_array[2] + theta * 
			parameter_array[3] ));
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown omega function type %ld.", function_type );
		break;
	}

	position_array[ omega_motor_index ] = omega;

	/******** Evaluate new position for insertion device gap **********/

	/* If the gap _is_ supposed to be linked to all the other motions,
	 * then read in the rest of the gap parameters.
	 */

	if ( gap_linked ) {
		/* We need to convert the requested theta value into 
		 * units of electron volts so that we can compute the
		 * new gap value from it.
		 */

		status = mx_get_double_variable_by_name( list_head,
				aps_10id_motor->d_spacing_record->name,
				&d_spacing );

		if ( status.code != MXE_SUCCESS )
			return status;

		angle_in_radians = theta * (aps_10id_motor->angle_scale);

		mono_energy = MX_HC
				/ ( 2.0 * d_spacing * sin(angle_in_radians));

		/* Get the requested offset for the undulator gap. */

		status = mx_get_double_variable_by_name( list_head,
				aps_10id_motor->gap_offset_record->name,
				&gap_offset );

		if ( status.code != MXE_SUCCESS )
			return status;

		new_gap_energy = mono_energy + gap_offset;

		/* Find out which harmonic of the undulator radiation
		 * that we want to use.
		 */

		status = mx_get_long_variable_by_name( list_head,
				aps_10id_motor->gap_harmonic_record->name,
				&gap_harmonic );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( gap_harmonic <= 0 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Requested gap harmonic value = %ld is illegal.  "
			"The gap harmonic must be greater than zero.",
				gap_harmonic );
		}

		new_gap_energy /= (double) gap_harmonic;

		position_array[ gap_energy_motor_index ] = new_gap_energy;
	}

	/********************** Now move the motors. **********************/

	if ( (function_type == MXF_APS_10ID_NO_DEPENDENCE)
				  && (gap_linked == FALSE) ) {

		real_theta_motor_record =
			motor_record_array[ real_theta_motor_index ];

		status = mx_motor_move_absolute( real_theta_motor_record,
							theta, 0 );
	} else {
		status = mx_motor_array_move_absolute( num_motors,
				motor_record_array, position_array, 0 );
	}

	return status;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_get_position()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *real_theta_motor_record;
	double theta;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	real_theta_motor_record = aps_10id_motor->real_theta_motor_record;

	if ( real_theta_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"real_theta_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_get_position( real_theta_motor_record, &theta );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor->raw_position.analog = theta;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_set_position()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *real_theta_motor_record;
	double theta;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed NULL." );
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	real_theta_motor_record = aps_10id_motor->real_theta_motor_record;

	if ( real_theta_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Dependent MX_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	theta = motor->raw_set_position.analog;

	status = mx_motor_set_position( real_theta_motor_record, theta );

	return status;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_soft_abort()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *real_theta_motor_record;
	MX_RECORD *omega_motor_record;
	MX_RECORD *gap_energy_motor_record;
	mx_status_type status1, status2, status3;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	real_theta_motor_record = aps_10id_motor->real_theta_motor_record;

	if ( real_theta_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"real_theta_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	omega_motor_record = aps_10id_motor->omega_motor_record;

	if ( omega_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"omega_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	gap_energy_motor_record = aps_10id_motor->gap_energy_motor_record;

	if ( gap_energy_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"gap_energy_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status1 = mx_motor_soft_abort( real_theta_motor_record );

	status2 = mx_motor_soft_abort( omega_motor_record );

	status3 = mx_motor_soft_abort( gap_energy_motor_record );

	if ( status1.code != MXE_SUCCESS ) {
		return status1;
	} else {
		if ( status2.code != MXE_SUCCESS ) {
			return status2;
		} else {
			return status3;
		}
	}
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_immediate_abort()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *real_theta_motor_record;
	MX_RECORD *omega_motor_record;
	MX_RECORD *gap_energy_motor_record;
	mx_status_type status1, status2, status3;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	real_theta_motor_record = aps_10id_motor->real_theta_motor_record;

	if ( real_theta_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"real_theta_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	omega_motor_record = aps_10id_motor->omega_motor_record;

	if ( omega_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"omega_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	gap_energy_motor_record = aps_10id_motor->gap_energy_motor_record;

	if ( gap_energy_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"gap_energy_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status1 = mx_motor_immediate_abort( real_theta_motor_record );

	status2 = mx_motor_immediate_abort( omega_motor_record );

	status3 = mx_motor_immediate_abort( gap_energy_motor_record );

	if ( status1.code != MXE_SUCCESS ) {
		return status1;
	} else {
		if ( status2.code != MXE_SUCCESS ) {
			return status2;
		} else {
			return status3;
		}
	}
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_positive_limit_hit()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *real_theta_motor_record;
	MX_RECORD *omega_motor_record;
	int limit_hit;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	real_theta_motor_record = aps_10id_motor->real_theta_motor_record;

	if ( real_theta_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"real_theta_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	omega_motor_record = aps_10id_motor->omega_motor_record;

	if ( omega_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"omega_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_positive_limit_hit(
				real_theta_motor_record, &limit_hit );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( limit_hit ) {
		motor->positive_limit_hit = TRUE;

		return MX_SUCCESSFUL_RESULT;
	} else {
		/* It is safest to check both limits, so that we are not
		 * dependent on future changes to the omega(theta) function.
		 */
		status = mx_motor_positive_limit_hit(
					omega_motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit ) {
			motor->positive_limit_hit = TRUE;

			return MX_SUCCESSFUL_RESULT;
		}

		status = mx_motor_negative_limit_hit(
					omega_motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	motor->positive_limit_hit = limit_hit;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_negative_limit_hit()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *real_theta_motor_record;
	MX_RECORD *omega_motor_record;
	int limit_hit;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	real_theta_motor_record = aps_10id_motor->real_theta_motor_record;

	if ( real_theta_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"real_theta_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	omega_motor_record = aps_10id_motor->omega_motor_record;

	if ( omega_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"omega_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_negative_limit_hit(
				real_theta_motor_record, &limit_hit );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( limit_hit ) {
		motor->negative_limit_hit = TRUE;

		return MX_SUCCESSFUL_RESULT;
	} else {
		/* It is safest to check both limits, so that we are not
		 * dependent on future changes to the omega(theta) function.
		 */
		status = mx_motor_positive_limit_hit(
					omega_motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit ) {
			motor->negative_limit_hit = TRUE;

			return MX_SUCCESSFUL_RESULT;
		}

		status = mx_motor_negative_limit_hit(
					omega_motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	motor->negative_limit_hit = limit_hit;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_10id_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_10id_motor_find_home_position()";

	MX_APS_10ID_MOTOR *aps_10id_motor;
	MX_RECORD *real_theta_motor_record;
	MX_RECORD *omega_motor_record;
	int direction;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_10id_motor = (MX_APS_10ID_MOTOR *)
				(motor->record->record_type_struct);

	if ( aps_10id_motor == (MX_APS_10ID_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_10ID_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	real_theta_motor_record = aps_10id_motor->real_theta_motor_record;

	if ( real_theta_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"real_theta_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	omega_motor_record = aps_10id_motor->omega_motor_record;

	if ( omega_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"omega_motor_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	direction = motor->home_search;

	status = mx_motor_find_home_position( real_theta_motor_record,
							direction );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_motor_find_home_position( omega_motor_record,
							- direction );

	return status;
}

