/*
 * Name:    d_net_amplifier.c
 *
 * Purpose: MX driver to control network amplifiers.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_amplifier.h"
#include "d_net_amplifier.h"

/* Initialize the amplifier driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_network_amplifier_record_function_list = {
	mxd_network_amplifier_initialize_type,
	mxd_network_amplifier_create_record_structures,
	mxd_network_amplifier_finish_record_initialization,
	mxd_network_amplifier_delete_record,
	NULL,
	mxd_network_amplifier_read_parms_from_hardware,
	mxd_network_amplifier_write_parms_to_hardware,
	mxd_network_amplifier_open,
	mxd_network_amplifier_close,
	NULL,
	mxd_network_amplifier_resynchronize
};

MX_AMPLIFIER_FUNCTION_LIST mxd_network_amplifier_amplifier_function_list = {
	mxd_network_amplifier_get_gain,
	mxd_network_amplifier_set_gain,
	mxd_network_amplifier_get_offset,
	mxd_network_amplifier_set_offset,
	mxd_network_amplifier_get_time_constant,
	mxd_network_amplifier_set_time_constant,
	mxd_network_amplifier_get_parameter,
	mx_amplifier_default_set_parameter_handler
};

/* Network amplifier data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_network_amplifier_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_AMPLIFIER_STANDARD_FIELDS,
	MXD_NETWORK_AMPLIFIER_STANDARD_FIELDS
};

long mxd_network_amplifier_num_record_fields
		= sizeof( mxd_network_amplifier_record_field_defaults )
		  / sizeof( mxd_network_amplifier_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_amplifier_rfield_def_ptr
			= &mxd_network_amplifier_record_field_defaults[0];

/*=======================================================================*/

/* This function is a utility function to consolidate all of the
 * pointer mangling that often has to happen at the beginning of an 
 * mxd_network_amplifier_...  function.  The parameter 'calling_fname'
 * is passed so that error messages will appear with the name of
 * the calling function.
 */

MX_EXPORT mx_status_type
mxd_network_amplifier_get_pointers( MX_AMPLIFIER *amplifier,
			MX_NETWORK_AMPLIFIER **network_amplifier,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_amplifier_get_pointers()";

	if ( amplifier == (MX_AMPLIFIER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The amplifier pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( amplifier->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_RECORD pointer for amplifier pointer passed by '%s' is NULL.",
			calling_fname );
	}

	if ( amplifier->record->mx_type != MXT_AMP_NETWORK ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The amplifier '%s' passed by '%s' is not a network amplifier.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			amplifier->record->name, calling_fname,
			amplifier->record->mx_superclass,
			amplifier->record->mx_class,
			amplifier->record->mx_type );
	}

	if ( network_amplifier != (MX_NETWORK_AMPLIFIER **) NULL ) {

		*network_amplifier = (MX_NETWORK_AMPLIFIER *)
				(amplifier->record->record_type_struct);

		if ( *network_amplifier == (MX_NETWORK_AMPLIFIER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_NETWORK_AMPLIFIER pointer for amplifier record '%s' passed by '%s' is NULL",
				amplifier->record->name, calling_fname );
		}
	}

	if ( (*network_amplifier)->server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX server record pointer for MX network amplifier '%s' is NULL.",
			amplifier->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/


MX_EXPORT mx_status_type
mxd_network_amplifier_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_network_amplifier_create_record_structures()";

	MX_AMPLIFIER *amplifier;
	MX_NETWORK_AMPLIFIER *network_amplifier;

	/* Allocate memory for the necessary structures. */

	amplifier = (MX_AMPLIFIER *) malloc( sizeof(MX_AMPLIFIER) );

	if ( amplifier == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_AMPLIFIER structure." );
	}

	network_amplifier = (MX_NETWORK_AMPLIFIER *)
				malloc( sizeof(MX_NETWORK_AMPLIFIER) );

	if ( network_amplifier == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_NETWORK_AMPLIFIER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = amplifier;
	record->record_type_struct = network_amplifier;
	record->class_specific_function_list
			= &mxd_network_amplifier_amplifier_function_list;

	amplifier->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_finish_record_initialization( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_delete_record( MX_RECORD *record )
{
	MX_NETWORK_AMPLIFIER *network_amplifier;

	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	network_amplifier = (MX_NETWORK_AMPLIFIER *) record->record_type_struct;

	if ( network_amplifier != NULL ) {
		free( network_amplifier );

		record->record_type_struct = NULL;
	}

	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxd_network_amplifier_resynchronize()";

	MX_AMPLIFIER *amplifier;
	MX_NETWORK_AMPLIFIER *network_amplifier;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int resynchronize;
	mx_status_type status;

	amplifier = (MX_AMPLIFIER *) record->record_class_struct;

	if ( amplifier == (MX_AMPLIFIER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_AMPLIFIER pointer for record '%s' is NULL.",
			record->name );
	}

	status = mxd_network_amplifier_get_pointers(
				amplifier, &network_amplifier, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf(rfname, "%s.resynchronize",
			network_amplifier->remote_record_name);

	resynchronize = 1;

	status = mx_put( network_amplifier->server_record,
				rfname, MXFT_INT, &resynchronize );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_get_gain( MX_AMPLIFIER *amplifier )
{
	const char fname[] = "mxd_network_amplifier_get_gain()";

	MX_NETWORK_AMPLIFIER *network_amplifier;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double gain;
	mx_status_type status;

	status = mxd_network_amplifier_get_pointers(
				amplifier, &network_amplifier, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.gain", network_amplifier->remote_record_name );

	status = mx_get( network_amplifier->server_record,
				rfname, MXFT_DOUBLE, &gain );

	amplifier->gain = gain;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_set_gain( MX_AMPLIFIER *amplifier )
{
	const char fname[] = "mxd_network_amplifier_set_gain()";

	MX_NETWORK_AMPLIFIER *network_amplifier;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double new_gain;
	mx_status_type status;

	status = mxd_network_amplifier_get_pointers(
				amplifier, &network_amplifier, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	new_gain = amplifier->gain;

	sprintf( rfname, "%s.gain", network_amplifier->remote_record_name );

	status = mx_put( network_amplifier->server_record,
				rfname, MXFT_DOUBLE, &new_gain );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_get_offset( MX_AMPLIFIER *amplifier )
{
	const char fname[] = "mxd_network_amplifier_get_offset()";

	MX_NETWORK_AMPLIFIER *network_amplifier;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double offset;
	mx_status_type status;

	status = mxd_network_amplifier_get_pointers(
				amplifier, &network_amplifier, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.offset", network_amplifier->remote_record_name );

	status = mx_get( network_amplifier->server_record,
				rfname, MXFT_DOUBLE, &offset );

	amplifier->offset = offset;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_set_offset( MX_AMPLIFIER *amplifier )
{
	const char fname[] = "mxd_network_amplifier_set_offset()";

	MX_NETWORK_AMPLIFIER *network_amplifier;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double new_offset;
	mx_status_type status;

	status = mxd_network_amplifier_get_pointers(
				amplifier, &network_amplifier, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	new_offset = amplifier->offset;

	sprintf( rfname, "%s.offset", network_amplifier->remote_record_name );

	status = mx_put( network_amplifier->server_record,
				rfname, MXFT_DOUBLE, &new_offset );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_get_time_constant( MX_AMPLIFIER *amplifier )
{
	const char fname[] = "mxd_network_amplifier_get_time_constant()";

	MX_NETWORK_AMPLIFIER *network_amplifier;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double time_constant;
	mx_status_type status;

	status = mxd_network_amplifier_get_pointers(
				amplifier, &network_amplifier, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.time_constant",
				network_amplifier->remote_record_name );

	status = mx_get( network_amplifier->server_record,
				rfname, MXFT_DOUBLE, &time_constant );

	amplifier->time_constant = time_constant;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_set_time_constant( MX_AMPLIFIER *amplifier )
{
	const char fname[] = "mxd_network_amplifier_set_time_constant()";

	MX_NETWORK_AMPLIFIER *network_amplifier;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double new_time_constant;
	mx_status_type status;

	status = mxd_network_amplifier_get_pointers(
				amplifier, &network_amplifier, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	new_time_constant = amplifier->time_constant;

	sprintf( rfname, "%s.time_constant",
				network_amplifier->remote_record_name );

	status = mx_put( network_amplifier->server_record,
				rfname, MXFT_DOUBLE, &new_time_constant );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_amplifier_get_parameter( MX_AMPLIFIER *amplifier )
{
	const char fname[] = "mxd_network_amplifier_get_parameter()";

	MX_NETWORK_AMPLIFIER *network_amplifier;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	mx_status_type status;

	status = mxd_network_amplifier_get_pointers(
				amplifier, &network_amplifier, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 2,
		("%s invoked for amplifier '%s' for parameter type '%s' (%d).",
		fname, amplifier->record->name,
		mx_get_field_label_string( amplifier->record,
						amplifier->parameter_type ),
		amplifier->parameter_type));

	switch( amplifier->parameter_type ) {
	case MXLV_AMP_GAIN_RANGE:
		sprintf( rfname, "%s.gain_range",
				network_amplifier->remote_record_name );

		dimension_array[0] = MX_AMPLIFIER_NUM_GAIN_RANGE_PARAMS;

		status = mx_get_array( network_amplifier->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				amplifier->gain_range );
		break;

	default:
		return mx_amplifier_default_get_parameter_handler( amplifier );
		break;
	}

	return status;
}

