/*
 * Name:    d_network_ainput.c
 *
 * Purpose: MX network analog input device driver.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_net.h"
#include "mx_analog_input.h"
#include "d_network_ainput.h"

MX_RECORD_FUNCTION_LIST mxd_network_ainput_record_function_list = {
	mxd_network_ainput_initialize_type,
	mxd_network_ainput_create_record_structures,
	mxd_network_ainput_finish_record_initialization,
	mxd_network_ainput_delete_record,
	NULL,
	mxd_network_ainput_read_parms_from_hardware,
	mxd_network_ainput_write_parms_to_hardware,
	mxd_network_ainput_open,
	mxd_network_ainput_close
};

MX_ANALOG_INPUT_FUNCTION_LIST mxd_network_ainput_analog_input_function_list = {
	mxd_network_ainput_read
};

MX_RECORD_FIELD_DEFAULTS mxd_network_ainput_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_DOUBLE_ANALOG_INPUT_STANDARD_FIELDS,
	MX_ANALOG_INPUT_STANDARD_FIELDS,
	MXD_NETWORK_AINPUT_STANDARD_FIELDS
};

long mxd_network_ainput_num_record_fields
		= sizeof( mxd_network_ainput_record_field_defaults )
			/ sizeof( mxd_network_ainput_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_ainput_rfield_def_ptr
			= &mxd_network_ainput_record_field_defaults[0];

/* ===== */

MX_EXPORT mx_status_type
mxd_network_ainput_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_create_record_structures( MX_RECORD *record )
{
        const char fname[] = "mxd_network_ainput_create_record_structures()";

        MX_ANALOG_INPUT *analog_input;
        MX_NETWORK_AINPUT *network_ainput;

        /* Allocate memory for the necessary structures. */

        analog_input = (MX_ANALOG_INPUT *) malloc(sizeof(MX_ANALOG_INPUT));

        if ( analog_input == (MX_ANALOG_INPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_ANALOG_INPUT structure." );
        }

        network_ainput = (MX_NETWORK_AINPUT *)
				malloc( sizeof(MX_NETWORK_AINPUT) );

        if ( network_ainput == (MX_NETWORK_AINPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_NETWORK_AINPUT structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = analog_input;
        record->record_type_struct = network_ainput;
        record->class_specific_function_list
			= &mxd_network_ainput_analog_input_function_list;

        analog_input->record = record;

	/* Raw analog input values are stored as doubles. */

	analog_input->subclass = MXT_AIN_DOUBLE;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_finish_record_initialization( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_read( MX_ANALOG_INPUT *ainput )
{
	const char fname[] = "mxd_network_ainput_read()";

	MX_NETWORK_AINPUT *network_ainput;
	char rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];
	long raw_value;
	mx_status_type status;

	network_ainput = (MX_NETWORK_AINPUT *)
				ainput->record->record_type_struct;

	if ( network_ainput == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_AINPUT pointer for analog input '%s' is NULL",
			ainput->record->name );
	}

	sprintf( rfname, "%s.raw_value", network_ainput->remote_record_name );

	status = mx_get( network_ainput->server_record,
				rfname, MXFT_DOUBLE, &raw_value );

	ainput->raw_value.double_value = raw_value;

	return status;
}

