/*
 * Name:    d_network_scaler.c
 *
 * Purpose: MX scaler driver to control MX network scalers.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_net.h"
#include "mx_scaler.h"
#include "d_network_scaler.h"

/* Initialize the scaler driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_network_scaler_record_function_list = {
	mxd_network_scaler_initialize_type,
	mxd_network_scaler_create_record_structures,
	mxd_network_scaler_finish_record_initialization,
	mxd_network_scaler_delete_record,
	NULL,
	mxd_network_scaler_read_parms_from_hardware,
	mxd_network_scaler_write_parms_to_hardware,
	mxd_network_scaler_open,
	mxd_network_scaler_close
};

MX_SCALER_FUNCTION_LIST mxd_network_scaler_scaler_function_list = {
	mxd_network_scaler_clear,
	mxd_network_scaler_overflow_set,
	mxd_network_scaler_read,
	mxd_network_scaler_read_and_clear,
	mxd_network_scaler_is_busy,
	mxd_network_scaler_start,
	mxd_network_scaler_stop,
	mxd_network_scaler_get_mode,
	mxd_network_scaler_set_mode
};

/* MX network scaler data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_network_scaler_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_SCALER_STANDARD_FIELDS,
	MXD_NETWORK_SCALER_STANDARD_FIELDS
};

long mxd_network_scaler_num_record_fields
		= sizeof( mxd_network_scaler_record_field_defaults )
		  / sizeof( mxd_network_scaler_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_scaler_rfield_def_ptr
			= &mxd_network_scaler_record_field_defaults[0];

/*=======================================================================*/

static mx_status_type
mxd_network_scaler_get_pointers( MX_SCALER *scaler,
			MX_NETWORK_SCALER **network_scaler,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_scaler_get_pointers()";

	if ( scaler == (MX_SCALER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The scaler pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( scaler->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_RECORD pointer for scaler pointer passed by '%s' is NULL.",
			calling_fname );
	}

	if ( scaler->record->mx_type != MXT_SCL_NETWORK ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The scaler '%s' passed by '%s' is not a network scaler.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			scaler->record->name, calling_fname,
			scaler->record->mx_superclass,
			scaler->record->mx_class,
			scaler->record->mx_type );
	}

	if ( network_scaler != (MX_NETWORK_SCALER **) NULL ) {

		*network_scaler = (MX_NETWORK_SCALER *)
				(scaler->record->record_type_struct);

		if ( *network_scaler == (MX_NETWORK_SCALER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_NETWORK_SCALER pointer for scaler record '%s' passed by '%s' is NULL",
				scaler->record->name, calling_fname );
		}
	}

	if ( (*network_scaler)->server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX server record pointer for MX network scaler '%s' is NULL.",
			scaler->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mxd_network_scaler_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_scaler_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_network_scaler_create_record_structures()";

	MX_SCALER *scaler;
	MX_NETWORK_SCALER *network_scaler;

	/* Allocate memory for the necessary structures. */

	scaler = (MX_SCALER *) malloc( sizeof(MX_SCALER) );

	if ( scaler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCALER structure." );
	}

	network_scaler = (MX_NETWORK_SCALER *)
				malloc( sizeof(MX_NETWORK_SCALER) );

	if ( network_scaler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_NETWORK_SCALER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = scaler;
	record->record_type_struct = network_scaler;
	record->class_specific_function_list
			= &mxd_network_scaler_scaler_function_list;

	scaler->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_scaler_finish_record_initialization( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_scaler_delete_record( MX_RECORD *record )
{
	MX_NETWORK_SCALER *network_scaler;

	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	network_scaler = (MX_NETWORK_SCALER *) record->record_type_struct;

	if ( network_scaler != NULL ) {
		free( network_scaler );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_scaler_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_scaler_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_scaler_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_scaler_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_scaler_clear( MX_SCALER *scaler )
{
	const char fname[] = "mxd_network_scaler_clear()";

	MX_NETWORK_SCALER *network_scaler;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long value;
	mx_status_type status;

	status = mxd_network_scaler_get_pointers(
				scaler, &network_scaler, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	value = 1;

	sprintf( rfname, "%s.clear", network_scaler->remote_record_name );

	status = mx_put( network_scaler->server_record,
				rfname, MXFT_INT, &value );

	scaler->value = 0L;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_scaler_overflow_set( MX_SCALER *scaler )
{
	const char fname[] = "mxd_network_scaler_overflow_set()";

	MX_NETWORK_SCALER *network_scaler;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int overflow_set;
	mx_status_type status;

	status = mxd_network_scaler_get_pointers(
				scaler, &network_scaler, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.overflow_set",
			network_scaler->remote_record_name );

	status = mx_get( network_scaler->server_record,
				rfname, MXFT_INT, &overflow_set );

	scaler->overflow_set = overflow_set;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_scaler_read( MX_SCALER *scaler )
{
	const char fname[] = "mxd_network_scaler_read()";

	MX_NETWORK_SCALER *network_scaler;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long value;
	mx_status_type status;

	status = mxd_network_scaler_get_pointers(
				scaler, &network_scaler, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.value", network_scaler->remote_record_name );

	status = mx_get( network_scaler->server_record,
				rfname, MXFT_LONG, &value );

	scaler->value = value;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_scaler_read_and_clear( MX_SCALER *scaler )
{
	mx_status_type status;

	status = mxd_network_scaler_read( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxd_network_scaler_clear( scaler );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_scaler_is_busy( MX_SCALER *scaler )
{
	const char fname[] = "mxd_network_scaler_is_busy()";

	MX_NETWORK_SCALER *network_scaler;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int busy;
	mx_status_type status;

	status = mxd_network_scaler_get_pointers(
				scaler, &network_scaler, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.busy",
			network_scaler->remote_record_name );

	status = mx_get( network_scaler->server_record,
				rfname, MXFT_INT, &busy );

	scaler->busy = busy;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_scaler_start( MX_SCALER *scaler )
{
	const char fname[] = "mxd_network_scaler_start()";

	MX_NETWORK_SCALER *network_scaler;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long value;
	mx_status_type status;

	status = mxd_network_scaler_get_pointers(
				scaler, &network_scaler, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.value", network_scaler->remote_record_name );

	value = scaler->value;

	status = mx_put( network_scaler->server_record,
				rfname, MXFT_LONG, &value );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_scaler_stop( MX_SCALER *scaler )
{
	const char fname[] = "mxd_network_scaler_stop()";

	MX_NETWORK_SCALER *network_scaler;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int stop;
	long value;
	mx_status_type status;

	status = mxd_network_scaler_get_pointers(
				scaler, &network_scaler, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.stop",
			network_scaler->remote_record_name );

	stop = 1;

	status = mx_put( network_scaler->server_record,
				rfname, MXFT_INT, &stop );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.value",
			network_scaler->remote_record_name );

	status = mx_get( network_scaler->server_record,
				rfname, MXFT_LONG, &value );

	scaler->value = value;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_scaler_get_mode( MX_SCALER *scaler )
{
	const char fname[] = "mxd_network_scaler_get_mode()";

	MX_NETWORK_SCALER *network_scaler;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int mode;
	mx_status_type status;

	status = mxd_network_scaler_get_pointers(
				scaler, &network_scaler, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.mode", network_scaler->remote_record_name );

	status = mx_get( network_scaler->server_record,
				rfname, MXFT_INT, &mode );

	scaler->mode = mode;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_scaler_set_mode( MX_SCALER *scaler )
{
	const char fname[] = "mxd_network_scaler_set_mode()";

	MX_NETWORK_SCALER *network_scaler;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int mode;
	mx_status_type status;

	status = mxd_network_scaler_get_pointers(
				scaler, &network_scaler, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	mode = scaler->mode;

	sprintf( rfname, "%s.mode", network_scaler->remote_record_name );

	status = mx_put( network_scaler->server_record,
				rfname, MXFT_INT, &mode );

	return status;
}

