/*
 * Name:    d_pmac.c
 *
 * Purpose: MX driver for Delta Tau PMAC controlled motors.
 *
 *          Note that this driver performs its work using the PMAC jog
 *          commands, rather than using a motion program.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "d_pmac.h"

/* ============ Motor channels ============ */

MX_RECORD_FUNCTION_LIST mxd_pmac_record_function_list = {
	mxd_pmac_initialize_type,
	mxd_pmac_create_record_structures,
	mxd_pmac_finish_record_initialization,
	mxd_pmac_delete_record,
	mxd_pmac_print_structure,
	mxd_pmac_read_parms_from_hardware,
	mxd_pmac_write_parms_to_hardware,
	mxd_pmac_open,
	mxd_pmac_close,
	NULL,
	mxd_pmac_resynchronize
};

MX_MOTOR_FUNCTION_LIST mxd_pmac_motor_function_list = {
	mxd_pmac_motor_is_busy,
	mxd_pmac_move_absolute,
	mxd_pmac_get_position,
	mxd_pmac_set_position,
	mxd_pmac_soft_abort,
	mxd_pmac_immediate_abort,
	mxd_pmac_positive_limit_hit,
	mxd_pmac_negative_limit_hit,
	mxd_pmac_find_home_position,
	mxd_pmac_constant_velocity_move,
	mxd_pmac_get_parameter,
	mxd_pmac_set_parameter
};

#define PMAC_DEBUG	FALSE

MX_RECORD_FIELD_DEFAULTS mxd_pmac_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_STEPPER_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_PMAC_STANDARD_FIELDS
};

long mxd_pmac_num_record_fields
		= sizeof( mxd_pmac_record_field_defaults )
			/ sizeof( mxd_pmac_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_pmac_rfield_def_ptr
			= &mxd_pmac_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_pmac_get_pointers( MX_MOTOR *motor,
			MX_PMAC_MOTOR **pmac_motor,
			const char *calling_fname )
{
	const char fname[] = "mxd_pmac_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( pmac_motor == (MX_PMAC_MOTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( *pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_PMAC_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_pmac_motor_get_pmac_pointer( MX_PMAC_MOTOR *pmac_motor,
			MX_PMAC **pmac,
			const char *calling_fname )
{
	const char fname[] = "mxd_pmac_motor_get_pmac_pointer()";

	MX_RECORD *pmac_record;

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( pmac == (MX_PMAC **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC pointer passed by '%s' was NULL.",
			calling_fname );
	}

	pmac_record = pmac_motor->pmac_record;

	if ( pmac_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The pmac_record pointer for record '%s' is NULL.",
			pmac_motor->record->name );
	}

	*pmac = (MX_PMAC *) pmac_record->record_type_struct;

	if ( *pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_PMAC pointer for record '%s' is NULL.",
			pmac_motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_create_record_structures()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) malloc( sizeof(MX_PMAC_MOTOR) );

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_PMAC_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = pmac_motor;
	record->class_specific_function_list
				= &mxd_pmac_motor_function_list;

	motor->record = record;
	pmac_motor->record = record;

	/* A PMAC motor is treated as an stepper motor. */

	motor->subclass = MXC_MTR_STEPPER;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_finish_record_initialization()";

	MX_PMAC_MOTOR *pmac_motor;
	double delay_seconds;
	mx_status_type mx_status;

	mx_status = mx_motor_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/*** Server minimum delay time between commands. ***/

	delay_seconds = 0.0;

	record->event_interval
		= mx_convert_seconds_to_clock_ticks( delay_seconds );

	pmac_motor = (MX_PMAC_MOTOR *) record->record_type_struct;

	if ( ( pmac_motor->motor_number < 1 )
				|| ( pmac_motor->motor_number > 32 ) )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Illegal PMAC motor number %d for record '%s'.  Allowed range is 1 to 32.",
			pmac_motor->motor_number, record->name );
	}

	pmac_motor->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_print_structure()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	long motor_steps;
	double position, backlash, negative_limit, positive_limit, deadband;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);

	fprintf(file, "  Motor type     = PMAC motor.\n\n");

	fprintf(file, "  name           = %s\n", record->name);
	fprintf(file, "  port name      = %s\n", pmac_motor->pmac_record->name);
	fprintf(file, "  motor number   = %d\n", pmac_motor->motor_number);

	mx_status = mx_motor_get_position_steps( record, &motor_steps );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
				record->name );
	}
	
	position = motor->offset + motor->scale
			* (double) motor_steps;
	
	fprintf(file, "  position       = %ld steps (%g %s)\n",
			motor_steps, position, motor->units);
	fprintf(file, "  scale          = %g %s per step.\n",
			motor->scale, motor->units);
	fprintf(file, "  offset         = %g %s.\n",
			motor->offset, motor->units);
	
	backlash = motor->scale
			* (double) (motor->raw_backlash_correction.stepper);
	
	fprintf(file, "  backlash       = %ld steps (%g %s)\n",
			motor->raw_backlash_correction.stepper,
			backlash, motor->units);
	
	negative_limit = motor->offset
	+ motor->scale * (double)(motor->raw_negative_limit.stepper);
	
	fprintf(file, "  negative limit = %ld steps (%g %s)\n",
			motor->raw_negative_limit.stepper,
			negative_limit, motor->units);

	positive_limit = motor->offset
	+ motor->scale * (double)(motor->raw_positive_limit.stepper);

	fprintf(file, "  positive limit = %ld steps (%g %s)\n",
			motor->raw_positive_limit.stepper,
			positive_limit, motor->units);

	deadband = motor->scale * (double) motor->deadband.stepper;

	fprintf(file, "  deadband       = %ld steps (%g %s)\n\n",
			motor->deadband.stepper,
			deadband, motor->units );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_read_parms_from_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_read_parms_from_hardware()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	double position;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Get the current position.  This has the side effect of
	 * updating the position value in the MX_MOTOR structure.
	 */

	mx_status = mx_motor_get_position( record, &position );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_resynchronize()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed was NULL." );
	}

	motor = (MX_MOTOR *) record->record_class_struct;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_resynchronize_record( pmac_motor->pmac_record );

	return mx_status;
}

/* ============ Motor specific functions ============ */

MX_EXPORT mx_status_type
mxd_pmac_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_motor_is_busy()";

	MX_PMAC_MOTOR *pmac_motor;
	char response[50];
	int length;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "?",
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	if ( (response[2] & 0x2) || (response[11] & 0x1) ) {
		motor->busy = FALSE;
	} else {
		motor->busy = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_move_absolute()";

	MX_PMAC_MOTOR *pmac_motor;
	MX_PMAC *pmac;
	char command[20];
	long motor_steps;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor_steps = motor->raw_destination.stepper;

	if ( ( (pmac->pmac_type & MX_PMAC_TYPE_TURBO) == 0 )
	  && ( pmac->major_version <= 1 )
	  && ( pmac->minor_version < 16 ) )
	{
		sprintf( command, "J=%ld", motor_steps );
	} else {
		sprintf( command, "J==%ld", motor_steps );
	}

	mx_status = mxd_pmac_jog_command( pmac_motor, command,
						NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_get_position_steps()";

	MX_PMAC_MOTOR *pmac_motor;
	char response[50];
	int num_tokens;
	double double_value;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "P",
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The PMAC may report values back in terms of fractions of
	 * an encoder count.
	 */

	num_tokens = sscanf( response, "%lg", &double_value );

	if ( num_tokens != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"No position value seen in response to 'P' command.  "
		"Response seen = '%s'", response );
	}

	/* We round this to the nearest encoder count. */

	motor->raw_position.stepper = mx_round( double_value );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_set_position()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( motor->raw_set_position.stepper != 0 ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Setting the position of motor '%s' to non-zero values is not supported.",
			motor->record->name );
	}

	mx_status = mxd_pmac_jog_command( pmac_motor, "HOMEZ",
						NULL, 0, PMAC_DEBUG );
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_soft_abort()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "J/", NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_immediate_abort()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "K", NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_positive_limit_hit()";

	MX_PMAC_MOTOR *pmac_motor;
	char response[50];
	int length;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "?",
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	if ( (response[0] & 0x2) == 0 ) {
		motor->positive_limit_hit = FALSE;
	} else {
		motor->positive_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_negative_limit_hit()";

	MX_PMAC_MOTOR *pmac_motor;
	char response[50];
	int length;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "?",
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	if ( (response[0] & 0x4) == 0 ) {
		motor->negative_limit_hit = FALSE;
	} else {
		motor->negative_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_find_home_position()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "HM", NULL, 0, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_constant_velocity_move()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( motor->constant_velocity_move > 0 ) {
		mx_status = mxd_pmac_jog_command( pmac_motor, "J+",
						NULL, 0, PMAC_DEBUG );
	} else
	if ( motor->constant_velocity_move < 0 ) {
		mx_status = mxd_pmac_jog_command( pmac_motor, "J-",
						NULL, 0, PMAC_DEBUG );
	} else {
		mx_status = MX_SUCCESSFUL_RESULT;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_get_parameter()";

	MX_PMAC_MOTOR *pmac_motor;
	double double_value;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		mx_status = mxd_pmac_get_motor_variable( pmac_motor, 22,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Reported by the controller in count/msec. */

		motor->raw_speed = 1000.0 * double_value;
		break;
	case MXLV_MTR_BASE_SPEED:
		/* The PMAC doesn't seem to have the concept of a base speed,
		 * since it is oriented to servo motors.
		 */

		motor->raw_base_speed = 0.0;

		break;
	case MXLV_MTR_ACCELERATION_PARAMETERS:
		mx_status = mxd_pmac_get_motor_variable( pmac_motor, 19,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		motor->acceleration_parameters[0] = double_value;

		mx_status = mxd_pmac_get_motor_variable( pmac_motor, 20,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		motor->acceleration_parameters[1] = double_value;

		mx_status = mxd_pmac_get_motor_variable( pmac_motor, 21,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		motor->acceleration_parameters[2] = double_value;

		motor->acceleration_parameters[3] = 0.0;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Parameter type %d is not supported by the driver for motor '%s'",
			motor->parameter_type, motor->record->name );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_set_parameter()";

	MX_PMAC_MOTOR *pmac_motor;
	double double_value;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		double_value = 0.001 * motor->raw_speed;

		mx_status = mxd_pmac_set_motor_variable( pmac_motor, 22,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	case MXLV_MTR_BASE_SPEED:
		/* The PMAC doesn't seem to have the concept of a base speed,
		 * since it is oriented to servo motors.
		 */

		break;
	case MXLV_MTR_ACCELERATION_PARAMETERS:
		double_value = motor->acceleration_parameters[0];

		mx_status = mxd_pmac_set_motor_variable( pmac_motor, 19,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		double_value = motor->acceleration_parameters[1];

		mx_status = mxd_pmac_set_motor_variable( pmac_motor, 20,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		double_value = motor->acceleration_parameters[2];

		mx_status = mxd_pmac_set_motor_variable( pmac_motor, 21,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_jog_command( MX_PMAC_MOTOR *pmac_motor,
			char *command,
			char *response,
			int response_buffer_length,
			int debug_flag )
{
	const char fname[] = "mxd_pmac_jog_command()";

	MX_PMAC *pmac;
	char command_buffer[100];
	char *ptr;
	mx_status_type mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( pmac->num_cards > 1 ) {
		sprintf( command_buffer, "@%x", pmac_motor->card_number );
	} else {
		strcpy( command_buffer, "" );
	}

	ptr = command_buffer + strlen( command_buffer );

	sprintf( ptr, "#%d%s", pmac_motor->motor_number, command );

	mx_status = mxi_pmac_command( pmac, command_buffer,
				response, response_buffer_length, debug_flag );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_get_motor_variable( MX_PMAC_MOTOR *pmac_motor,
				int variable_number,
				int variable_type,
				void *variable_ptr,
				int debug_flag )
{
	const char fname[] = "mxd_pmac_get_motor_variable()";

	MX_PMAC *pmac;
	char command_buffer[100];
	char response[100];
	char *ptr;
	int num_items;
	long long_value;
	long *long_ptr;
	double double_value;
	double *double_ptr;
	mx_status_type mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( pmac->num_cards > 1 ) {
		sprintf( command_buffer, "@%x", pmac_motor->card_number );
	} else {
		strcpy( command_buffer, "" );
	}

	ptr = command_buffer + strlen( command_buffer );

	if ( pmac->pmac_type & MX_PMAC_TYPE_TURBO ) {
		sprintf( ptr, "I02%d02%d",
				pmac_motor->motor_number, variable_number );
	} else {
		sprintf( ptr, "I%d%02d",
				pmac_motor->motor_number, variable_number );
	}

	mx_status = mxi_pmac_command( pmac, command_buffer,
				response, sizeof response, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( variable_type ) {
	case MXFT_LONG:
		num_items = sscanf( response, "%ld", &long_value );

		if ( num_items != 1 ) {
			return mx_error( MXE_UNPARSEABLE_STRING, fname,
"Cannot parse response to command '%s' as a long integer.  Response = '%s'",
				command_buffer, response );
		}
		long_ptr = (long *) variable_ptr;

		*long_ptr = long_value;
		break;
	case MXFT_DOUBLE:
		num_items = sscanf( response, "%lg", &double_value );

		if ( num_items != 1 ) {
			return mx_error( MXE_UNPARSEABLE_STRING, fname,
"Cannot parse response to command '%s' as a double.  Response = '%s'",
				command_buffer, response );
		}
		double_ptr = (double *) variable_ptr;

		*double_ptr = double_value;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Only MXFT_LONG and MXFT_DOUBLE are supported." );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_set_motor_variable( MX_PMAC_MOTOR *pmac_motor,
				int variable_number,
				int variable_type,
				void *variable_ptr,
				int debug_flag )
{
	const char fname[] = "mxd_pmac_set_motor_variable()";

	MX_PMAC *pmac;
	char command_buffer[100];
	char response[100];
	char *ptr;
	long *long_ptr;
	double *double_ptr;
	mx_status_type mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( pmac->num_cards > 1 ) {
		sprintf( command_buffer, "@%x", pmac_motor->card_number );
	} else {
		strcpy( command_buffer, "" );
	}

	ptr = command_buffer + strlen( command_buffer );

	switch( variable_type ) {
	case MXFT_LONG:
		long_ptr = (long *) variable_ptr;

		if ( pmac->pmac_type & MX_PMAC_TYPE_TURBO ) {
			sprintf( ptr, "I%02d%02d=%ld",
				pmac_motor->motor_number,
				variable_number,
				*long_ptr );
		} else {
			sprintf( ptr, "I%d%02d=%ld",
				pmac_motor->motor_number,
				variable_number,
				*long_ptr );
		}
		break;
	case MXFT_DOUBLE:
		double_ptr = (double *) variable_ptr;

		if ( pmac->pmac_type & MX_PMAC_TYPE_TURBO ) {
			sprintf( ptr, "I%02d%02d=%g",
				pmac_motor->motor_number,
				variable_number,
				*double_ptr );
		} else {
			sprintf( ptr, "I%d%02d=%g",
				pmac_motor->motor_number,
				variable_number,
				*double_ptr );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Only MXFT_LONG and MXFT_DOUBLE are supported." );
		break;
	}

	mx_status = mxi_pmac_command( pmac, command_buffer,
				response, sizeof response, debug_flag );

	return mx_status;
}

