/*
 * Name:    d_sis3801.c
 *
 * Purpose: MX MCS driver for the Struck SIS3801 multichannel scaler.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_driver.h"
#include "mx_clock.h"
#include "mx_vme.h"
#include "mx_mcs.h"

#include "d_sis3801.h"

MX_RECORD_FUNCTION_LIST mxd_sis3801_record_function_list = {
	mxd_sis3801_initialize_type,
	mxd_sis3801_create_record_structures,
	mxd_sis3801_finish_record_initialization,
	mxd_sis3801_delete_record,
	NULL,
	mxd_sis3801_read_parms_from_hardware,
	mxd_sis3801_write_parms_to_hardware,
	mxd_sis3801_open,
	mxd_sis3801_close,
	NULL,
	mxd_sis3801_resynchronize
};

MX_MCS_FUNCTION_LIST mxd_sis3801_mcs_function_list = {
	mxd_sis3801_start,
	mxd_sis3801_stop,
	mxd_sis3801_clear,
	mxd_sis3801_busy,
	mxd_sis3801_read_all,
	mxd_sis3801_read_scaler,
	mxd_sis3801_read_measurement,
	NULL,
	mxd_sis3801_get_parameter,
	mxd_sis3801_set_parameter
};

MX_RECORD_FIELD_DEFAULTS mxd_sis3801_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCS_STANDARD_FIELDS,
	MXD_SIS3801_STANDARD_FIELDS
};

long mxd_sis3801_num_record_fields
		= sizeof( mxd_sis3801_record_field_defaults )
			/ sizeof( mxd_sis3801_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_sis3801_rfield_def_ptr
			= &mxd_sis3801_record_field_defaults[0];

/* --- */

static mx_status_type
mxd_sis3801_get_pointers( MX_MCS *mcs,
			MX_SIS3801 **sis3801,
			const char *calling_fname )
{
	const char fname[] = "mxd_sis3801_get_pointers()";

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( sis3801 == (MX_SIS3801 **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_SIS3801 pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mcs->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_RECORD pointer for the MX_MCS structure passed by '%s' is NULL.",
			calling_fname );
	}

	*sis3801 = (MX_SIS3801 *) (mcs->record->record_type_struct);

	if ( *sis3801 == (MX_SIS3801 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SIS3801 pointer for record '%s' is NULL.",
			mcs->record->name );
	}
	return MX_SUCCESSFUL_RESULT;
}

/* ====== */

MX_EXPORT mx_status_type
mxd_sis3801_initialize_type( long record_type )
{
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	long num_record_fields;
	long maximum_num_scalers_varargs_cookie;
	long maximum_num_measurements_varargs_cookie;
	mx_status_type status;

	status = mx_mcs_initialize_type( record_type,
				&num_record_fields,
				&record_field_defaults,
				&maximum_num_scalers_varargs_cookie,
				&maximum_num_measurements_varargs_cookie );

	return status;
}

MX_EXPORT mx_status_type
mxd_sis3801_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_sis3801_create_record_structures()";

	MX_MCS *mcs;
	MX_SIS3801 *sis3801;

	mcs = (MX_MCS *) malloc( sizeof(MX_MCS) );

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_MCS structure." );
	}

	sis3801 = (MX_SIS3801 *) malloc( sizeof(MX_SIS3801) );

	if ( sis3801 == (MX_SIS3801 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_SIS3801 structure." );
	}

	record->record_class_struct = mcs;
	record->record_type_struct = sis3801;

	record->class_specific_function_list = &mxd_sis3801_mcs_function_list;

	mcs->record = record;
	sis3801->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_finish_record_initialization( MX_RECORD *record )
{
	MX_MCS *mcs;
	mx_status_type mx_status;

	mx_status = mx_mcs_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mcs = (MX_MCS *) record->record_class_struct;

	mcs->first_scaler_number = 1;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_open( MX_RECORD *record )
{
	const char fname[] = "mxd_sis3801_open()";

	MX_MCS *mcs;
	MX_SIS3801 *sis3801;
	mx_uint32_type module_id_register;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	mcs = (MX_MCS *) record->record_class_struct;

	mx_status = mxd_sis3801_get_pointers( mcs, &sis3801, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Parse the address mode string. */

	mx_status = mx_vme_parse_address_mode( sis3801->address_mode_name,
						&(sis3801->address_mode) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Reset the SIS3801. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_RESET_REG,
				1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Read the module identification register. */

	mx_status = mx_vme_in32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
		sis3801->base_address + MX_SIS3801_MODULE_ID_IRQ_CONTROL_REG,
				&module_id_register );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sis3801->module_id = ( module_id_register >> 16 ) & 0xffff;
	sis3801->firmware_version = ( module_id_register >> 12 ) & 0xf;

	MX_DEBUG( 2,("%s: module id = %#lx, firmware version = %lu",
		fname, sis3801->module_id, sis3801->firmware_version));

	if ( sis3801->sis3801_flags & MXF_SIS3801_USE_REFERENCE_PULSER ) {

		/* Enable reference pulses. */

		mx_status = mx_vme_out32( sis3801->vme_record,
					sis3801->crate_number,
					sis3801->address_mode,
	sis3801->base_address + MX_SIS3801_ENABLE_REFERENCE_PULSE_CHANNEL_1,
					1 );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* Set the measurement finish time to the current time. */

	sis3801->finish_time = mx_current_clock_tick();

	mcs->measurement_time = 0.0;

	/* We assume there are no counts remaining to be read in the FIFO. */

	sis3801->counts_available_in_fifo = FALSE;

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_resynchronize( MX_RECORD *record )
{
	return mxd_sis3801_open( record );
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_sis3801_start( MX_MCS *mcs )
{
	const char fname[] = "mxd_sis3801_start()";

	MX_SIS3801 *sis3801;
	MX_CLOCK_TICK start_time, finish_time, measurement_ticks;
	double total_measurement_time;
	mx_uint32_type prescale_factor, control_register;
	mx_status_type mx_status;

	mx_status = mxd_sis3801_get_pointers( mcs, &sis3801, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, mcs->record->name));

	/* Enable or disable the internal 10 MHz clock. */

	if ( sis3801->sis3801_flags & MXF_SIS3801_USE_EXTERNAL_CHANNEL_ADVANCE )
	{
		control_register = MXF_SIS3801_DISABLE_10MHZ_TO_LNE_PRESCALER;
	} else {
		control_register = MXF_SIS3801_ENABLE_10MHZ_TO_LNE_PRESCALER;
	}

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_CONTROL_REG,
				control_register );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Disable the LNE prescaler. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_CONTROL_REG,
				MXF_SIS3801_DISABLE_LNE_PRESCALER );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Set the prescale factor. */

	if ( sis3801->sis3801_flags & MXF_SIS3801_USE_EXTERNAL_CHANNEL_ADVANCE )
	{
		prescale_factor = mx_round( sis3801->external_prescale );
	} else {
		prescale_factor = mx_round( MX_SIS3801_10MHZ_INTERNAL_CLOCK
					* mcs->measurement_time );
	}

	prescale_factor -= 1;

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
			sis3801->base_address + MX_SIS3801_PRESCALE_FACTOR_REG,
				prescale_factor );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Enable the LNE prescaler. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_CONTROL_REG,
				MXF_SIS3801_ENABLE_LNE_PRESCALER );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Enable the next clock logic. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
		sis3801->base_address + MX_SIS3801_ENABLE_NEXT_CLOCK_LOGIC,
				1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Enable counting. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_CONTROL_REG,
			MXF_SIS3801_CLEAR_SOFTWARE_DISABLE_COUNTING_BIT );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sis3801->counts_available_in_fifo = TRUE;

#if 1
	/* Send a clock pulse. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
			sis3801->base_address + MX_SIS3801_SEND_NEXT_CLOCK,
				1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#endif

	/* Record the time when the measurement should be finished. */

	total_measurement_time = mcs->measurement_time
				* (double) mcs->current_num_measurements;

	/* Add one extra measurement time if we are skipping the
	 * first measurement.
	 */

	if ( (sis3801->sis3801_flags
		& MXF_SIS3801_DO_NOT_SKIP_FIRST_MEASUREMENT) == 0 ) 
	{
		total_measurement_time += ( 1.0 * mcs->measurement_time );
	}

	/* Give a little extra delay time to try to avoid reading out the
	 * FIFO before all the measurements have been taken.
	 */

	total_measurement_time += 0.02;		/* Delay time in seconds */

	measurement_ticks
		= mx_convert_seconds_to_clock_ticks( total_measurement_time );

	start_time = mx_current_clock_tick();

	finish_time = mx_add_clock_ticks( start_time, measurement_ticks );

	MX_DEBUG( 2,("%s: measurement_time = %g, measurement_ticks = (%lu,%lu)",
		fname, total_measurement_time, measurement_ticks.high_order,
			(unsigned long) measurement_ticks.low_order ));

	MX_DEBUG( 2,("%s: start_time = (%lu,%lu), finish_time = (%lu,%lu)",
		fname, start_time.high_order,
		(unsigned long) start_time.low_order,
		finish_time.high_order,
		(unsigned long) finish_time.low_order));

	sis3801->finish_time = finish_time;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_stop( MX_MCS *mcs )
{
	const char fname[] = "mxd_sis3801_stop()";

	MX_SIS3801 *sis3801;
	mx_status_type mx_status;

	mx_status = mxd_sis3801_get_pointers( mcs, &sis3801, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, mcs->record->name));

	/* Disable counting. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_CONTROL_REG,
				MXF_SIS3801_SET_SOFTWARE_DISABLE_COUNTING_BIT );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Disable next clock logic. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
		sis3801->base_address + MX_SIS3801_DISABLE_NEXT_CLOCK_LOGIC,
				1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Set the measurement finish time to the current time. */

	sis3801->finish_time = mx_current_clock_tick();

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_clear( MX_MCS *mcs )
{
	const char fname[] = "mxd_sis3801_clear()";

	MX_SIS3801 *sis3801;
	unsigned long i, j;
	mx_status_type mx_status;

	mx_status = mxd_sis3801_get_pointers( mcs, &sis3801, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, mcs->record->name));

	/* Clear the local data array. */

	for ( i = 0; i < mcs->maximum_num_scalers; i++ ) {
		for ( j = 0; j < mcs->maximum_num_measurements; j++ ) {
			mcs->data_array[i][j] = 0;
		}
	}

	/* Clear the FIFO. */

	mx_status = mx_vme_out32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
			sis3801->base_address + MX_SIS3801_CLEAR_FIFO_REG,
				1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sis3801->counts_available_in_fifo = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_busy( MX_MCS *mcs )
{
	const char fname[] = "mxd_sis3801_busy()";

	MX_SIS3801 *sis3801;
	MX_CLOCK_TICK current_time;
	mx_status_type mx_status;

	mx_status = mxd_sis3801_get_pointers( mcs, &sis3801, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, mcs->record->name));

	/* See if we have reached the expected finish time of the measurement.*/

	current_time = mx_current_clock_tick();

	if ( mx_compare_clock_ticks( current_time, sis3801->finish_time ) >= 0 )
	{
		mcs->busy = FALSE;
	} else {
		mcs->busy = TRUE;
	}

	if ( mcs->busy == FALSE ) {
		MX_DEBUG( 2,
	("%s: busy = %d, current_time = (%lu,%lu), finish_time = (%lu,%lu)",
		fname, mcs->busy, current_time.high_order,
		(unsigned long) current_time.low_order,
		sis3801->finish_time.high_order,
		(unsigned long) sis3801->finish_time.low_order));

		/* Stop any counting that may still be going on. */

		mx_status = mxd_sis3801_stop( mcs );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_read_all( MX_MCS *mcs )
{
	const char fname[] = "mxd_sis3801_read_all()";

	MX_SIS3801 *sis3801;
	long i, j, num_measurements, loop_start;
	mx_uint32_type fifo_value, status_register;
	mx_status_type mx_status;

	mx_status = mxd_sis3801_get_pointers( mcs, &sis3801, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, mcs->record->name));

	/* Don't try to read new counts if there aren't any to read. */

	if ( sis3801->counts_available_in_fifo == FALSE ) {
		return MX_SUCCESSFUL_RESULT;
	}

	sis3801->counts_available_in_fifo = FALSE;

	if (sis3801->sis3801_flags & MXF_SIS3801_DO_NOT_SKIP_FIRST_MEASUREMENT)
	{
		loop_start = 0;
	} else {
		/* The real first measurement is intended to be thrown away,
		 * which is why the loop starts at j == -1, rather than 0.
		 * We throw away the first measurement since we do not know
		 * precisely where the internal clock was in its cycle when
		 * we enabled counting.
		 */
		loop_start = -1;
	}

	num_measurements = (long) mcs->current_num_measurements;

	for ( j = loop_start; j < num_measurements; j++ ) {

		for ( i = 0; i < mcs->current_num_scalers; i++ ) {

			/* Check the FIFO status. */

			mx_status = mx_vme_in32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_STATUS_REG,
				&status_register );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			MX_DEBUG( 2,("%s: status_register = %#lx",
				fname, (unsigned long) status_register));

			/* If the FIFO is full, the SIS3801 will have locked
			 * up and needs to be reset.
			 */

			if ( status_register & MXF_SIS3801_FIFO_FLAG_FULL ) {
				return mx_error( MXE_LIMIT_WAS_EXCEEDED, fname,
"The MCS '%s' FIFO reports it is full at measurement %ld, scaler %ld.  "
"This is an unrecoverable error and will require a software reset of the MCS.",
					mcs->record->name, j, i );
			}

			/* If the FIFO is empty, we have run out of data
			 * before we should have.
			 */

			if ( status_register & MXF_SIS3801_FIFO_FLAG_EMPTY ) {
				MX_DEBUG(-2,(
	"%s: The MCS '%s' FIFO ran out of data at measurement %ld, scaler %ld "
	"when there should have been %lu measurements for the %lu scalers.",
		fname, mcs->record->name,
		j, i, mcs->current_num_measurements, mcs->current_num_scalers));

				return MX_SUCCESSFUL_RESULT;
			}

			/* Read the next data value. */

			mx_status = mx_vme_in32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_READ_FIFO,
				&fifo_value );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			if ( j >= 0 ) {
				mcs->data_array[i][j] = (long) fifo_value;
			}
		}
	}

	MX_DEBUG( 2,
	("%s: %lu measurements for %lu scalers were successfully read.",
	fname, mcs->current_num_measurements, mcs->current_num_scalers));

#if 0
	/* Check the FIFO status one more time. */

	mx_status = mx_vme_in32( sis3801->vme_record,
				sis3801->crate_number,
				sis3801->address_mode,
				sis3801->base_address + MX_SIS3801_STATUS_REG,
				&status_register );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_read_scaler( MX_MCS *mcs )
{
	mx_status_type mx_status;

	/* Read the data from the FIFO. */

	mx_status = mxd_sis3801_read_all( mcs );

	/* Since mcs->data_array has the scaler number as its leading
	 * array index, the data is already laid out in mcs->data_array
	 * in such a way that it can be passed back by mx_mcs_read_scaler()
	 * to the application program by merely passing back a pointer
	 * to mcs->data_array[scaler_number][0].
	 *
	 * Thus, we do not need to copy any data, since it is already laid
	 * out in memory the way we want it to be.
	 */

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_sis3801_read_measurement( MX_MCS *mcs )
{
	const char fname[] = "mxd_sis3801_read_measurement()";

	MX_SIS3801 *sis3801;
	unsigned long i, j;
	mx_status_type mx_status;

	mx_status = mxd_sis3801_get_pointers( mcs, &sis3801, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, mcs->record->name));

	/* Read the data from the FIFO. */

	mx_status = mxd_sis3801_read_all( mcs );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Copy the data to the measurement_data array. */

	j = mcs->measurement_index;

	for ( i = 0; i < mcs->current_num_scalers; i++ ) {
		mcs->measurement_data[i] = mcs->data_array[i][j];
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3801_get_parameter( MX_MCS *mcs )
{
	const char fname[] = "mxd_sis3801_get_parameter()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_sis3801_set_parameter( MX_MCS *mcs )
{
	const char fname[] = "mxd_sis3801_set_parameter()";

	MX_SIS3801 *sis3801;
	mx_status_type mx_status;

	mx_status = mxd_sis3801_get_pointers( mcs, &sis3801, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for MCS '%s', parameter type '%s' (%d)",
		fname, mcs->record->name,
		mx_get_field_label_string( mcs->record, mcs->parameter_type ),
		mcs->parameter_type));

	if ( mcs->parameter_type == MXLV_MCS_MODE ) {

	} else if ( mcs->parameter_type == MXLV_MCS_MEASUREMENT_TIME ) {

	} else if ( mcs->parameter_type == MXLV_MCS_CURRENT_NUM_MEASUREMENTS ){

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			mcs->parameter_type );
	}
	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

