/*
 * Name:    i_epics_vme.c
 *
 * Purpose: MX driver for VME access via the EPICS VME record written by
 *          Mark Rivers of the University of Chicago.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_driver.h"
#include "mx_vme.h"
#include "i_epics_vme.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

#include "mx_epics.h"

MX_RECORD_FUNCTION_LIST mxi_epics_vme_record_function_list = {
	mxi_epics_vme_initialize_type,
	mxi_epics_vme_create_record_structures,
	mxi_epics_vme_finish_record_initialization,
	mxi_epics_vme_delete_record,
	NULL,
	mxi_epics_vme_read_parms_from_hardware,
	mxi_epics_vme_write_parms_to_hardware,
	mxi_epics_vme_open,
	mxi_epics_vme_close,
	NULL,
	mxi_epics_vme_resynchronize
};

MX_VME_FUNCTION_LIST mxi_epics_vme_vme_function_list = {
	mxi_epics_vme_input,
	mxi_epics_vme_output,
	mxi_epics_vme_multi_input,
	mxi_epics_vme_multi_output,
	mxi_epics_vme_get_parameter,
	mxi_epics_vme_set_parameter
};

MX_RECORD_FIELD_DEFAULTS mxi_epics_vme_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS
};

long mxi_epics_vme_num_record_fields
		= sizeof( mxi_epics_vme_record_field_defaults )
			/ sizeof( mxi_epics_vme_record_field_defaults[0]);

MX_RECORD_FIELD_DEFAULTS *mxi_epics_vme_rfield_def_ptr
		= &mxi_epics_vme_record_field_defaults[0];

/*--*/

#define MX_EPICS_VME_DEBUG	TRUE

/*--*/

#if MX_EPICS_VME_DEBUG

/* MX_EPICS_VME_DEBUG_WARNING_SUPPRESS is used to suppress uninitialized
 * variable warnings from GCC.
 */

#if 0
#  define MX_EPICS_VME_DEBUG_WARNING_SUPPRESS \
	do { \
		uint8_ptr = NULL; uint16_ptr = NULL; uint32_ptr = NULL; \
	} while(0)
#endif

#  define MX_EPICS_VME_DEBUG_PRINT \
	do { \
		unsigned long value = 0; \
                                     \
		switch( vme->data_size ) { \
		case MXF_VME_D8:  value = (unsigned long) *uint8_ptr;  break; \
		case MXF_VME_D16: value = (unsigned long) *uint16_ptr; break; \
		case MXF_VME_D32: value = (unsigned long) *uint32_ptr; break; \
		} \
		MX_DEBUG(-2,("%s: A%lu, D%lu, addr = %#lx, value = %#lx", \
			fname, \
			(unsigned long) vme->address_mode, \
			(unsigned long) vme->data_size, \
			(unsigned long) vme->address, \
			(unsigned long) value )); \
	} while(0)

#  define MX_EPICS_VME_MULTI_DEBUG_PRINT \
	do { \
		unsigned long value = 0; \
                                     \
		switch( vme->data_size ) { \
		case MXF_VME_D8:  value = (unsigned long) *uint8_ptr;  break; \
		case MXF_VME_D16: value = (unsigned long) *uint16_ptr; break; \
		case MXF_VME_D32: value = (unsigned long) *uint32_ptr; break; \
		} \
	MX_DEBUG(-2,("%s: A%lu, D%lu, addr = %#lx, value = %#lx (%lu values)", \
			fname, \
			(unsigned long) vme->address_mode, \
			(unsigned long) vme->data_size, \
			(unsigned long) vme->address, \
			(unsigned long) value, \
			vme->num_values )); \
	} while(0)
#else

#  define MX_EPICS_VME_DEBUG_WARNING_SUPPRESS

#  define MX_EPICS_VME_DEBUG_PRINT

#  define MX_EPICS_VME_MULTI_DEBUG_PRINT

#endif

/*==========================*/

static mx_status_type
mxi_epics_vme_transfer_data( MX_VME *vme,
				MX_EPICS_VME *epics_vme,
				int direction )
{
	const char fname[] = "mxi_epics_vme_transfer_data()";

	char pvname[80];
	long i, epics_address_mode, epics_data_size, epics_num_values;
	long address_increment, epics_direction, epics_address;
	long *long_data_pointer;
	mx_uint8_type *uint8_ptr;
	mx_uint16_type *uint16_ptr;
	mx_status_type mx_status;

	/* Object if we are asked to transfer more data than the EPICS
	 * side is prepared for.
	 */

	if ( vme->num_values > epics_vme->max_epics_values ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Attempt to transfer %lu data values via EPICS VME record '%s' failed."
	"  EPICS record '%s' only allows up to %ld values to be transferred.",
			vme->num_values, vme->record->name,
			epics_vme->epics_record_name,
			epics_vme->max_epics_values );
	}

	/* Set the VME addressing mode. */

	if ( vme->address_mode != epics_vme->old_address_mode ) {

		switch( vme->address_mode ) {
		case MXF_VME_A16:
			epics_address_mode = 0;
			break;
		case MXF_VME_A24:
			epics_address_mode = 1;
			break;
		case MXF_VME_A32:
			epics_address_mode = 2;
			break;
		default:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal VME address mode %lu for EPICS VME record '%s'.  "
		"The allowed values are 16, 24, and 32.",
				vme->address_mode, vme->record->name );
			break;
		}
					
		sprintf( pvname, "%s.AMOD", epics_vme->epics_record_name );

		mx_status = mx_caput(pvname, ezcaLong, 1, &epics_address_mode);

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		epics_vme->old_address_mode = vme->address_mode;
	}

	/* Set the VME data size. */

	if ( vme->data_size != epics_vme->old_data_size ) {

		switch( vme->data_size ) {
		case MXF_VME_D8:
			epics_data_size = 0;
			break;
		case MXF_VME_D16:
			epics_data_size = 1;
			break;
		case MXF_VME_D32:
			epics_data_size = 2;
			break;
		default:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal VME data size %lu for EPICS VME record '%s'.  "
		"The allowed values are 8, 16, and 32.",
				vme->data_size, vme->record->name );
			break;
		}
					
		sprintf( pvname, "%s.DSIZ", epics_vme->epics_record_name );

		mx_status = mx_caput(pvname, ezcaLong, 1, &epics_data_size);

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		epics_vme->old_data_size = vme->data_size;
	}

	/* Set the number of values to transfer. */

	if ( vme->num_values != epics_vme->old_num_values ) {

		epics_num_values = vme->num_values;

		sprintf( pvname, "%s.NUSE", epics_vme->epics_record_name );

		mx_status = mx_caput( pvname, ezcaLong, 1, &epics_num_values );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		epics_vme->old_num_values = vme->num_values;
	}

	/* Set the address increment. */

	if ( direction == MXF_EPICS_VME_OUTPUT ) {
		address_increment = vme->write_address_increment;

	} else if ( direction == MXF_EPICS_VME_INPUT ) {
		address_increment = vme->read_address_increment;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal direction flag %d for EPICS VME record '%s'.  "
		"The allowed values are %d and %d.",
			direction, vme->record->name,
			MXF_EPICS_VME_OUTPUT, MXF_EPICS_VME_INPUT );
	}

	if ( address_increment != epics_vme->old_address_increment ) {

		if ( ( address_increment < 0 ) || ( address_increment > 4 ) )
		{
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal address increment %ld for EPICS VME record '%s'.  "
		"The allowed values are in the range (0-4).",
			address_increment, vme->record->name );
		}

		sprintf( pvname, "%s.AINC", epics_vme->epics_record_name );

		mx_status = mx_caput( pvname, ezcaLong, 1, &address_increment );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		epics_vme->old_address_increment = address_increment;
	}

	/* Set the data transfer direction. */

	if ( direction != epics_vme->old_direction ) {

		epics_direction = direction;

		sprintf( pvname, "%s.RDWT", epics_vme->epics_record_name );

		mx_status = mx_caput( pvname, ezcaLong, 1, &epics_direction );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		epics_vme->old_direction = direction;
	}

	/* Set the VME address. */

	if ( vme->address != epics_vme->old_address ) {

		epics_address = vme->address;

		sprintf( pvname, "%s.ADDR", epics_vme->epics_record_name );

		mx_status = mx_caput( pvname, ezcaLong, 1, &epics_address );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		epics_vme->old_address = vme->address;
	}

	/* Suppress a bogus uninitialized pointer warning from GCC. */

	long_data_pointer = NULL;

	/* If we are sending data to the VME crate, copy any 8 or 16 bit
	 * data to the temporary long buffer.
	 */

	if ( direction == MXF_EPICS_VME_OUTPUT ) {
		switch ( vme->data_size ) {
		case MXF_VME_D8:
			uint8_ptr = (mx_uint8_type *) vme->data_pointer;

			for ( i = 0; i < vme->num_values; i++ ) {
				epics_vme->temp_buffer[i]
					= (long) uint8_ptr[i];
			}
			long_data_pointer = epics_vme->temp_buffer;
			break;
		case MXF_VME_D16:
			uint16_ptr = (mx_uint16_type *) vme->data_pointer;

			for ( i = 0; i < vme->num_values; i++ ) {
				epics_vme->temp_buffer[i]
					= (long) uint16_ptr[i];
			}
			long_data_pointer = epics_vme->temp_buffer;
			break;
		case MXF_VME_D32:
			/* No copying is necessary for D32 data. */

			long_data_pointer = (long *) vme->data_pointer;
			break;
		}
	} else {
		long_data_pointer = (long *) vme->data_pointer;
	}

	/* Now we are finally ready to transfer the data. */

	sprintf( pvname, "%s.VAL", epics_vme->epics_record_name );

	if ( direction == MXF_EPICS_VME_OUTPUT ) {
		mx_status = mx_caput( pvname, ezcaLong,
					vme->num_values, long_data_pointer );
	} else {
		mx_status = mx_caget( pvname, ezcaLong,
					vme->num_values, long_data_pointer );
	}

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Check to see if there were any errors in the transfer. */

	sprintf( pvname, "%s.SARR", epics_vme->epics_record_name );

	mx_status = mx_caget( pvname, ezcaByte, vme->num_values, 
					epics_vme->status_array );

	for ( i = 0; i < vme->num_values; i++ ) {
		if ( epics_vme->status_array[i] != 0 ) {
			if ( direction == MXF_EPICS_VME_INPUT ) {
				return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Error reading array element %ld (D%lu) from address A%lu %lu "
		"for EPICS VME record '%s'.", i, vme->data_size,
					vme->address_mode, vme->address,
					epics_vme->record->name );
			} else {
				return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Error writing array element %ld (D%lu) to address A%lu %lu "
		"for EPICS VME record '%s'.", i, vme->data_size,
					vme->address_mode, vme->address,
					epics_vme->record->name );
			}
		}
	}

	/* If we are receiving data from the VME crate, copy any 8 or 16 bit
	 * data from the temporary long buffer.
	 */

	if ( direction == MXF_EPICS_VME_INPUT ) {
		switch ( vme->data_size ) {
		case MXF_VME_D8:
			uint8_ptr = (mx_uint8_type *) vme->data_pointer;

			for ( i = 0; i < vme->num_values; i++ ) {
				uint8_ptr[i] = (mx_uint8_type)
						epics_vme->temp_buffer[i];
			}
			break;
		case MXF_VME_D16:
			uint16_ptr = (mx_uint16_type *) vme->data_pointer;

			for ( i = 0; i < vme->num_values; i++ ) {
				uint16_ptr[i] = (mx_uint16_type)
						epics_vme->temp_buffer[i];
			}
			break;
		case MXF_VME_D32:
			/* No copying is necessary for D32 data. */

			break;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxi_epics_vme_get_pointers( MX_VME *vme,
				MX_EPICS_VME **epics_vme,
				const char calling_fname[] )
{
	const char fname[] = "mxi_epics_vme_get_pointers()";

	if ( vme == (MX_VME *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_VME pointer passed by '%s' is NULL.",
			calling_fname );
	}
	if ( epics_vme == (MX_EPICS_VME **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_VME pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*epics_vme = (MX_EPICS_VME *) vme->record->record_type_struct;

	if ( *epics_vme == (MX_EPICS_VME *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_EPICS_VME pointer for record '%s' is NULL.",
				vme->record->name );
	}
	return MX_SUCCESSFUL_RESULT;
}

/*---*/

MX_EXPORT mx_status_type
mxi_epics_vme_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_epics_vme_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_epics_vme_create_record_structures()";

	MX_VME *vme;
	MX_EPICS_VME *epics_vme;

	/* Allocate memory for the necessary structures. */

	vme = (MX_VME *) malloc( sizeof(MX_VME) );

	if ( vme == (MX_VME *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_VME structure." );
	}

	epics_vme = (MX_EPICS_VME *) malloc( sizeof(MX_EPICS_VME) );

	if ( epics_vme == (MX_EPICS_VME *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_VME structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = vme;
	record->record_type_struct = epics_vme;
	record->class_specific_function_list = &mxi_epics_vme_vme_function_list;

	vme->record = record;
	epics_vme->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_epics_vme_finish_record_initialization( MX_RECORD *record )
{
	return mx_vme_finish_record_initialization( record );
}

MX_EXPORT mx_status_type
mxi_epics_vme_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_epics_vme_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_epics_vme_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_epics_vme_open( MX_RECORD *record )
{
	const char fname[] = "mxi_epics_vme_open()";

	MX_VME *vme;
	MX_EPICS_VME *epics_vme;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	vme = (MX_VME *) record->record_class_struct;

	mx_status = mxi_epics_vme_get_pointers( vme, &epics_vme, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	epics_vme->old_address_mode = MXF_VME_A16;
	epics_vme->old_data_size = MXF_VME_D16;
	epics_vme->old_num_values = 1;
	epics_vme->old_address = 0;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_epics_vme_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_epics_vme_resynchronize( MX_RECORD *record )
{
	return mxi_epics_vme_open( record );
}

/*===============*/

MX_EXPORT mx_status_type
mxi_epics_vme_input( MX_VME *vme )
{
	const char fname[] = "mxi_epics_vme_input()";

	MX_EPICS_VME *epics_vme;
	mx_status_type mx_status;

	mx_status = mxi_epics_vme_get_pointers( vme, &epics_vme, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_epics_vme_transfer_data( vme,
					epics_vme, MXF_EPICS_VME_INPUT );

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_epics_vme_output( MX_VME *vme )
{
	const char fname[] = "mxi_epics_vme_output()";

	MX_EPICS_VME *epics_vme;
	mx_status_type mx_status;

	mx_status = mxi_epics_vme_get_pointers( vme, &epics_vme, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_epics_vme_transfer_data( vme,
					epics_vme, MXF_EPICS_VME_OUTPUT );

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_epics_vme_multi_input( MX_VME *vme )
{
	const char fname[] = "mxi_epics_vme_input()";

	MX_EPICS_VME *epics_vme;
	mx_status_type mx_status;

	mx_status = mxi_epics_vme_get_pointers( vme, &epics_vme, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_epics_vme_transfer_data( vme,
					epics_vme, MXF_EPICS_VME_INPUT );

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_epics_vme_multi_output( MX_VME *vme )
{
	const char fname[] = "mxi_epics_vme_output()";

	MX_EPICS_VME *epics_vme;
	mx_status_type mx_status;

	mx_status = mxi_epics_vme_get_pointers( vme, &epics_vme, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_epics_vme_transfer_data( vme,
					epics_vme, MXF_EPICS_VME_OUTPUT );

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_epics_vme_get_parameter( MX_VME *vme )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_epics_vme_set_parameter( MX_VME *vme )
{
	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */
