/*
 * Name:    i_vme58.c
 *
 * Purpose: MX driver for Oregon Microsystems VME58 motor controllers.
 *          This driver uses the ESRF Linux device driver for m68k Linux
 *          running on a Motorola 680x0 VME system controller.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2000-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_LINUX_VME58

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "omslib.h"

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_alloc.h"
#include "mx_generic.h"
#include "i_vme58.h"

MX_RECORD_FUNCTION_LIST mxi_vme58_record_function_list = {
	mxi_vme58_initialize_type,
	mxi_vme58_create_record_structures,
	mxi_vme58_finish_record_initialization,
	mxi_vme58_delete_record,
	NULL,
	mxi_vme58_read_parms_from_hardware,
	mxi_vme58_write_parms_to_hardware,
	mxi_vme58_open,
	mxi_vme58_close,
	NULL,
	mxi_vme58_resynchronize,
};

MX_GENERIC_FUNCTION_LIST mxi_vme58_generic_function_list = {
	mxi_vme58_getchar,
	mxi_vme58_putchar,
	mxi_vme58_read,
	mxi_vme58_write,
	mxi_vme58_input_is_available,
	mxi_vme58_discard_unread_input,
	mxi_vme58_discard_unwritten_output
};

MX_RECORD_FIELD_DEFAULTS mxi_vme58_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_VME58_STANDARD_FIELDS
};

long mxi_vme58_num_record_fields
		= sizeof( mxi_vme58_record_field_defaults )
			/ sizeof( mxi_vme58_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_vme58_rfield_def_ptr
			= &mxi_vme58_record_field_defaults[0];

#define MXI_VME58_DEBUG		FALSE

/* A private function for the use of the driver. */

static mx_status_type
mxi_vme58_get_pointers( MX_GENERIC *generic,
			MX_VME58 **vme58,
			const char *calling_fname )
{
	const char fname[] = "mxi_vme58_get_pointers()";

	MX_RECORD *vme58_record;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_GENERIC pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( vme58 == (MX_VME58 **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_VME58 pointer passed by '%s' was NULL.",
			calling_fname );
	}

	vme58_record = generic->record;

	if ( vme58_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The vme58_record pointer for the "
			"MX_GENERIC pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*vme58 = (MX_VME58 *) vme58_record->record_type_struct;

	if ( *vme58 == (MX_VME58 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_VME58 pointer for record '%s' is NULL.",
			vme58_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_vme58_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_vme58_create_record_structures()";

	MX_GENERIC *generic;
	MX_VME58 *vme58;

	/* Allocate memory for the necessary structures. */

	generic = (MX_GENERIC *) malloc( sizeof(MX_GENERIC) );

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_GENERIC structure." );
	}

	vme58 = (MX_VME58 *) malloc( sizeof(MX_VME58) );

	if ( vme58 == (MX_VME58 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_VME58 structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = generic;
	record->record_type_struct = vme58;
	record->class_specific_function_list = &mxi_vme58_generic_function_list;

	generic->record = record;
	vme58->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_finish_record_initialization( MX_RECORD *record )
{
	MX_VME58 *vme58;
	int i;

	vme58 = (MX_VME58 *) record->record_type_struct;

	/* Mark the Linux device file as being closed. */

	vme58->device_fd = -1;

	/* Initialize the motor array. */

	for ( i = 0; i < MX_MAX_VME58_AXES; i++ ) {
		vme58->motor_array[i] = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_delete_record( MX_RECORD *record )
{
	MX_VME58 *vme58;

        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }

	vme58 = (MX_VME58 *) record->record_type_struct;

        if ( vme58 != NULL ) {
                free( vme58 );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_open( MX_RECORD *record )
{
	const char fname[] = "mxi_vme58_open()";

	MX_VME58 *vme58;
	int saved_errno;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	vme58 = (MX_VME58 *) (record->record_type_struct);

	if ( vme58 == (MX_VME58 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_VME58 pointer for record '%s' is NULL.",
		record->name);
	}

	/* Try to open the Linux device file. */

	vme58->device_fd = open( vme58->device_name, O_RDWR );

	if ( vme58->device_fd == -1 ) {
		saved_errno = errno;

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Cannot open the VME58 device file '%s'.  Reason = '%s'.",
			vme58->device_name, strerror( saved_errno ) );
	}

	/* Synchronize with the controller. */

	status = mxi_vme58_resynchronize( record );

	return status;
}

MX_EXPORT mx_status_type
mxi_vme58_close( MX_RECORD *record )
{
	const char fname[] = "mxi_vme58_close()";

	MX_VME58 *vme58;
	int close_status, saved_errno;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	vme58 = (MX_VME58 *) (record->record_type_struct);

	if ( vme58 == (MX_VME58 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_VME58 pointer for record '%s' is NULL.",
		record->name);
	}

	if ( vme58->device_fd >= 0 ) {
		close_status = close( vme58->device_fd );

		if ( close_status != 0 ) {
			saved_errno = errno;

			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Cannot close the VME58 device file '%s'.  Reason = '%s'.",
			vme58->device_name, strerror( saved_errno ) );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxi_vme58_resynchronize()";

	MX_GENERIC *generic;
	MX_VME58 *vme58;
	char response[80];
	char type_string[80];
	char *version_ptr, *ptr;
	int num_items;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	generic = (MX_GENERIC *) record->record_class_struct;

	status = mxi_vme58_get_pointers( generic, &vme58, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Verify that the controller is there by asking it for its
	 * version number.
	 */

	status = mxi_vme58_command( vme58, "WY", response, sizeof response,
						MXI_VME58_DEBUG);
	
	switch( status.code ) {
	case MXE_SUCCESS:
		break;
	case MXE_INTERFACE_IO_ERROR:
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"Attempt to communicate with the VME58 controller '%s' failed.",
			vme58->device_name );
		break;
	default:
		return status;
		break;
	}

	/* Start with some conservative defaults. */

	vme58->controller_type = MXT_VME58_UNKNOWN;
	vme58->num_axes = 1;

	/* The first two characters in the response should an LF followed
	 * by a CR.
	 */

	if ( (response[0] != MX_LF) || (response[1] != MX_CR) ) {

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"VME58 interface '%s' did not return a response to the WY command.  "
"Instead we received '%s'.", record->name, response );
	}

	version_ptr = response + 2;

	if ( strncmp( version_ptr, "VME58", 5 ) != 0 )
	{
		/* Don't know what type of controller this is. */

		(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
"VME58 interface '%s' had an unrecognized response to the WY command.  "
"WY response = '%s'", record->name, version_ptr );

	} else {
		/* Find the hyphen in the response string.  This is just
		 * before the start of the controller type.
		 */

		ptr = strchr( version_ptr, '-' );

		if ( ptr == NULL ) {

			(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
"Cannot determine the particular controller type for VME interface '%s'.  "
"WY response = '%s'", record->name, version_ptr );

		} else {
			ptr++;		/* Skip over the hyphen. */

			num_items = sscanf( ptr, "%s", type_string );

			if ( num_items != 1 ) {

				(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
"Cannot find the particular controller type for VME interface '%s'.  "
"WY response = '%s'", record->name, version_ptr );

			} else {
				if ( strcmp( type_string, "8" ) == 0 ) {
					vme58->controller_type = MXT_VME58_8;
					vme58->num_axes = 8;
				} else
				if ( strcmp( type_string, "4E" ) == 0 ) {
					vme58->controller_type = MXT_VME58_4E;
					vme58->num_axes = 4;
				} else {
					(void) mx_error(MXE_UNPARSEABLE_STRING,
						fname,
"Unrecognized controller type '%s' in WY response for VME interface '%s'.  "
"WY response = '%s'", type_string, record->name, version_ptr );
				}
			}
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_getchar( MX_GENERIC *generic, char *c, int flags )
{
	const char fname[] = "mxi_vme58_getchar()";

	return mx_error( MXE_UNSUPPORTED, fname,
		"Single character I/O is not supported for the VME58." );
}

MX_EXPORT mx_status_type
mxi_vme58_putchar( MX_GENERIC *generic, char c, int flags )
{
	const char fname[] = "mxi_vme58_putchar()";

	return mx_error( MXE_UNSUPPORTED, fname,
		"Single character I/O is not supported for the VME58." );
}

MX_EXPORT mx_status_type
mxi_vme58_read( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_vme58_read()";

	return mx_error( MXE_UNSUPPORTED, fname,
		"Reading partial lines is not supported for the VME58." );
}

MX_EXPORT mx_status_type
mxi_vme58_write( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_vme58_write()";

	return mx_error( MXE_UNSUPPORTED, fname,
		"Writing partial lines is not supported for the VME58." );
}

MX_EXPORT mx_status_type
mxi_vme58_input_is_available( MX_GENERIC *generic, int *input_available )
{
	const char fname[] = "mxi_vme58_input_is_available()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"Input is available only in the context of sending a command.");
}

MX_EXPORT mx_status_type
mxi_vme58_discard_unread_input( MX_GENERIC *generic, int debug_flag )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_vme58_discard_unwritten_output( MX_GENERIC *generic, int debug_flag )
{
	return MX_SUCCESSFUL_RESULT;
}

/* === Functions specific to this driver. === */

static char *OMS_Error_Strings[] = {
	"No error!",
	"Invalid function",
	"Multiple axes specified",
	"Axis is busy",
	"Non-implemented function",
	"Timeout",
	"No axes specified",
	"Bad axis specification",
	"DONE interrupt pending",
	"Response overload",
	"String too long",
	"Bad position data from board",
	"Insufficient output buffer space",
	"Interrupt queued",
	"Bad response to status request",
	"Slip detected",
	"Limit detected",
	"Command error",
};

MX_EXPORT mx_status_type
mxi_vme58_command( MX_VME58 *vme58,
		char *command, char *response, int response_buffer_length,
		int debug_flag )
{
	const char fname[] = "mxi_vme58_command()";

	int oms_status;
	OMS_req_t request;

	if ( vme58 == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_VME58 pointer passed was NULL." );
	}
	if ( command == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"'command' buffer pointer passed was NULL.  No command sent.");
	}
	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: sending '%s'", fname, command ));
	}

	/* Set up the request structure */

	memset( &request, 0, sizeof(request) );

	request.cmdstr = command;
	request.cmdlen = strlen( command );

	/* Send the command. */

	if ( response == NULL ) {

		/* No response is expected. */

		oms_status = ioctl( vme58->device_fd, OMS_CMD_RET, &request );

	} else {

		/* A response is expected. */

		request.rspstr = response;
		request.rsplen = response_buffer_length;

		oms_status = ioctl( vme58->device_fd, OMS_CMD_RESP, &request );
	}

	if ( oms_status != 0 ) {

		if ( oms_status >= OMS_ERR_BASE_VALUE &&
			oms_status <= OMS_ERR_BASE_VALUE + OMS_ERR_RANGE )
		{
			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"An OMS error occurred for command '%s' to controller '%s'.  Error = '%s'.",
				command, vme58->device_name,
				OMS_Error_Strings[ oms_status
						- OMS_ERR_BASE_VALUE ] );
		} else {
			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"An operating system error occurred for command '%s' to controller '%s'.  "
"Error = '%s'.", command, vme58->device_name, strerror( oms_status ) );
		}
	}

	if ( response && debug_flag ) {
		MX_DEBUG(-2,("%s: received '%s'", fname, response ));
	}
	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_LINUX_VME58 */
