/*
 * Name:    mx_net.c
 *
 * Purpose: Client side part of MX network protocol.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifndef OS_MPW
#include <sys/types.h>
#endif

#include "mx_osdef.h"
#include "mx_util.h"
#include "mx_types.h"
#include "mx_record.h"
#include "mx_socket.h"
#include "mx_net.h"

#if ( USE_TCPIP == 0 )
#     define htonl(x) (x)
#     define ntohl(x) (x)
#endif

#define NETWORK_DEBUG	FALSE

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_network_receive_message( MX_RECORD *server_record,
				unsigned long buffer_length, void *buffer )
{
	const char fname[] = "mx_network_receive_message()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	mx_status_type ( *fptr ) ( MX_NETWORK_SERVER *, unsigned long, void * );
	mx_status_type status;

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Record pointer passed was NULL." );
	}
	if ( buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Message buffer pointer passed for record '%s' was NULL.",
			server_record->name );
	}

	server = (MX_NETWORK_SERVER *) (server_record->record_class_struct);

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	function_list = (MX_NETWORK_SERVER_FUNCTION_LIST *)
			(server_record->class_specific_function_list);

	if ( function_list == (MX_NETWORK_SERVER_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SERVER_FUNCTION_LIST pointer for server record '%s' is NULL.",
			server_record->name );
	}

	fptr = function_list->receive_message;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"receive_message function pointer for server record '%s' is NULL.",
			server_record->name );
	}

	status = ( *fptr ) ( server, buffer_length, buffer );

	return status;
}

MX_EXPORT mx_status_type
mx_network_send_message( MX_RECORD *server_record, void *buffer )
{
	const char fname[] = "mx_network_send_message()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	mx_status_type ( *fptr ) ( MX_NETWORK_SERVER *, void * );
	mx_status_type status;

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Record pointer passed was NULL." );
	}
	if ( buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Message buffer pointer passed for record '%s' was NULL.",
			server_record->name );
	}

	server = (MX_NETWORK_SERVER *) (server_record->record_class_struct);

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	function_list = (MX_NETWORK_SERVER_FUNCTION_LIST *)
			(server_record->class_specific_function_list);

	if ( function_list == (MX_NETWORK_SERVER_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SERVER_FUNCTION_LIST pointer for server record '%s' is NULL.",
			server_record->name );
	}

	fptr = function_list->send_message;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"send_message function pointer for server record '%s' is NULL.",
			server_record->name );
	}

	status = ( *fptr ) ( server, buffer );

	return status;
}

MX_EXPORT void
mx_network_display_message_buffer( void *buffer_ptr )
{
	const char fname[] = "mx_network_display_message_buffer()";

	mx_uint32_type *header;
	unsigned char *buffer, *message;
	mx_uint32_type magic_number, header_length, message_length;
	mx_uint32_type message_type, status_code, i;
	unsigned char c;

	if ( buffer_ptr == NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
			"Buffer pointer passed was NULL." );
		return;
	}

	buffer = buffer_ptr;
	header = buffer_ptr;

	magic_number   = ntohl( header[ MX_NETWORK_MAGIC ] );
	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code    = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	message = buffer + header_length;

	/**** Print out the header. ****/

	/* Header magic number */

	mx_info( "Magic number:     0x%8lx   (%2x %2x %2x %2x)",
	  (long) magic_number, buffer[0], buffer[1], buffer[2], buffer[3] );

	if ( magic_number != MX_NETWORK_MAGIC_VALUE ) {
		(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Invalid magic number 0x%lx in network buffer header.",
			(long) magic_number );
		return;
	}

	/* Header length */

	i = MX_NETWORK_HEADER_LENGTH * sizeof( mx_uint32_type );

	mx_info( "Header length:  %12lu   (%2x %2x %2x %2x)",
		header_length, buffer[i+0], buffer[i+1],
			buffer[i+2], buffer[i+3] );

	if ( header_length < MX_NETWORK_HEADER_LENGTH_VALUE ) {
		mx_info( "*** Header length %lu was unexpectedly short. ***",
			header_length );
		return;
	}

	/* Message length */

	i = MX_NETWORK_MESSAGE_LENGTH * sizeof( mx_uint32_type );

	mx_info( "Message length: %12lu   (%2x %2x %2x %2x)",
		message_length, buffer[i+0], buffer[i+1],
			buffer[i+2], buffer[i+3] );

	/* Message type */

	i = MX_NETWORK_MESSAGE_TYPE * sizeof( mx_uint32_type );

	mx_info( "Message type:   %12lu   (%2x %2x %2x %2x)",
		message_type, buffer[i+0], buffer[i+1],
			buffer[i+2], buffer[i+3] );

	/* Status code */

	i = MX_NETWORK_STATUS_CODE * sizeof( mx_uint32_type );

	mx_info( "Status code:    %12lu   (%2x %2x %2x %2x)",
		status_code, buffer[i+0], buffer[i+1],
			buffer[i+2], buffer[i+3] );

	/* Remaining part of header. */

	if ( header_length > MX_NETWORK_HEADER_LENGTH_VALUE ) {
		mx_info( "Remaining header bytes:" );

		for ( i += sizeof( mx_uint32_type ); i < header_length; i++ ) {
			c = buffer[i];

			if ( isprint(c) ) {
				mx_info("  buffer[%lu]  = %02x (%c)",
						i, c, c);
			} else {
				mx_info("  buffer[%lu]  = %02x", i, c);
			}
		}
	}

	/* Print out the message body. */

	mx_info( "Message contents:" );

	for ( i = 0; i < message_length; i++ ) {
		c = message[i];

		if ( isprint(c) ) {
			mx_info("  message[%lu] = %02x (%c)", i, c, c);
		} else {
			mx_info("  message[%lu] = %02x", i, c);
		}
	}
	return;
}

/* ====================================================================== */

static mx_status_type
mx_ascii_field_get_parameters( MX_RECORD *server_record,
			MX_RECORD_FIELD *local_field,
			long *datatype,
			long *num_dimensions,
			long **dimension,
			MX_NETWORK_SERVER **server,
			MX_NETWORK_SERVER_FUNCTION_LIST **function_list,
			const char *calling_fname )
{
	const char fname[] = "mx_ascii_field_get_parameters()";

	if ( server_record == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_record pointer passed to '%s' is NULL.",
			calling_fname );
	}
	if ( local_field == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"local_field pointer passed to '%s' is NULL.",
			calling_fname );
	}
	*datatype = local_field->datatype;

	switch( *datatype ) {
	case MXFT_STRING:
	case MXFT_CHAR:
	case MXFT_UCHAR:
	case MXFT_SHORT:
	case MXFT_USHORT:
	case MXFT_INT:
	case MXFT_UINT:
	case MXFT_LONG:
	case MXFT_ULONG:
	case MXFT_FLOAT:
	case MXFT_DOUBLE:
	case MXFT_HEX:
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Data type %ld passed to '%s' is not a legal network datatype.",
			*datatype, calling_fname );
		break;
	}

	*num_dimensions = local_field->num_dimensions;
	*dimension = local_field->dimension;
	*server = (MX_NETWORK_SERVER *) (server_record->record_class_struct);
	*function_list = (MX_NETWORK_SERVER_FUNCTION_LIST *)
				(server_record->class_specific_function_list);

	return MX_SUCCESSFUL_RESULT;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_get_array_ascii( MX_RECORD *server_record,
		char *remote_record_field_name,
		long datatype,
		long num_dimensions,
		long *dimension,
		void *value_ptr )
{
	const char fname[] = "mx_get_array_ascii()";

	MX_RECORD_FIELD local_temp_record_field;
	long *dimension_array;
	long local_dimension_array[MXU_FIELD_MAX_DIMENSIONS];
	size_t data_element_size[MXU_FIELD_MAX_DIMENSIONS];
	mx_status_type status;

	if ( dimension != NULL ) {
		dimension_array = dimension;

	} else if ( num_dimensions == 0 ) {
		dimension_array = local_dimension_array;
		dimension_array[0] = 0;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"dimension array pointer is NULL, but num_dimensions (%ld) is greater than 0",
			num_dimensions );
	}

	/* Setting the first element of data_element_size to be zero causes
	 * mx_construct_temp_record_field() to use a default data element
	 * size array.
	 */

	data_element_size[0] = 0L;

	/* Setup a temporary record field to be used by
	 * mx_get_field_array_ascii().
	 */

	status = mx_construct_temp_record_field( &local_temp_record_field,
			datatype, num_dimensions, dimension_array,
			data_element_size, value_ptr );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Send the request to the server. */

	status = mx_get_field_array_ascii( server_record,
			remote_record_field_name,
			&local_temp_record_field,
			value_ptr );

	return status;
}

MX_EXPORT mx_status_type
mx_put_array_ascii( MX_RECORD *server_record,
		char *remote_record_field_name,
		long datatype,
		long num_dimensions,
		long *dimension,
		void *value_ptr )
{
	const char fname[] = "mx_put_array_ascii()";

	MX_RECORD_FIELD local_temp_record_field;
	long *dimension_array;
	long local_dimension_array[MXU_FIELD_MAX_DIMENSIONS];
	size_t data_element_size[MXU_FIELD_MAX_DIMENSIONS];
	mx_status_type status;

	if ( dimension != NULL ) {
		dimension_array = dimension;

	} else if ( num_dimensions == 0 ) {
		dimension_array = local_dimension_array;
		dimension_array[0] = 0;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"dimension array pointer is NULL, but num_dimensions (%ld) is greater than 0",
			num_dimensions );
	}

	/* Setting the first element of data_element_size to be zero causes
	 * mx_construct_temp_record_field() to use a default data element
	 * size array.
	 */

	data_element_size[0] = 0L;

	/* Setup a temporary record field to be used by
	 * mx_put_field_array_ascii().
	 */

	status = mx_construct_temp_record_field( &local_temp_record_field,
			datatype, num_dimensions, dimension_array,
			data_element_size, value_ptr );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Send the request to the server. */

	status = mx_put_field_array_ascii( server_record,
			remote_record_field_name,
			&local_temp_record_field,
			value_ptr );

	return status;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_get_field_array_ascii( MX_RECORD *server_record,
			char *remote_record_field_name,
			MX_RECORD_FIELD *local_field,
			void *value_ptr )
{
	const char fname[] = "mx_get_field_array_ascii()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	MX_RECORD_FIELD_PARSE_STATUS temp_parse_status;
	mx_status_type ( *token_parser )
		(void *, char *, MX_RECORD *, MX_RECORD_FIELD_PARSE_STATUS *);
	long datatype, num_dimensions, *dimension;

	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header;
	char *buffer;
	char *message;
	mx_uint32_type header_length, message_length;
	mx_uint32_type message_type, status_code;
	mx_status_type status;
	char token_buffer[500];

	static char separators[] = MX_RECORD_FIELD_SEPARATORS;

	status = mx_ascii_field_get_parameters( server_record, local_field,
			&datatype, &num_dimensions, &dimension,
			&server, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 2,
	("%s: server_record = '%s', remote_record_field_name = '%s'",
		fname, server_record->name, remote_record_field_name));

	status = mx_get_token_parser( datatype, &token_parser );

	if ( status.code != MXE_SUCCESS )
		return status;

	/*********** Send a 'get array ascii' command. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->header[0]);
	buffer = &(aligned_buffer->buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_MESSAGE_TYPE]
				= htonl( MX_NETMSG_GET_ARRAY_ASCII );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;

	mx_strncpy( message, remote_record_field_name,
				MXU_RECORD_FIELD_NAME_LENGTH );

	message_length = strlen( message ) + 1;

	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( message_length );

#if NETWORK_DEBUG
	mx_network_display_message_buffer( buffer );
#endif

	status = mx_network_send_message( server_record, buffer );

	if ( status.code != MXE_SUCCESS )
		return status;

	/************* Wait for the response. **************/

	status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

	if ( status.code != MXE_SUCCESS )
		return status;

#if NETWORK_DEBUG
	mx_network_display_message_buffer( buffer );
#endif

	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code    = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	if ( message_type != mx_server_response(MX_NETMSG_GET_ARRAY_ASCII) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {
			return mx_error( (long)status_code, fname,
				"%s", message );
		}

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_GET_ARRAY_ASCII_RESPONSE.  "
"Instead it was of type = %ld", message_type );
	}

	message = buffer + header_length;

	/* If the remote command failed, the message field will include
	 * the text of the error message rather than the array data
	 * we wanted.
	 */

	if ( status_code != MXE_SUCCESS ) {
		return mx_error( (long)status_code, fname, message );
	}

	/************ Parse the tokens that were returned. ***************/

	temp_parse_status.description = message;
	temp_parse_status.separators = separators;
	temp_parse_status.num_separators = strlen(separators);
	temp_parse_status.start_of_trailing_whitespace = NULL;

	if ( (num_dimensions == 0) ||
		((datatype == MXFT_STRING) && (num_dimensions == 1)) ) {

		status = mx_get_next_record_token( &temp_parse_status,
				token_buffer, sizeof(token_buffer) );

		if ( status.code != MXE_SUCCESS )
			return status;

		status = ( *token_parser ) ( value_ptr,
				token_buffer, NULL, &temp_parse_status );
	} else {
		status = mx_parse_array_description( value_ptr,
				num_dimensions - 1, NULL, local_field,
				&temp_parse_status, token_parser );
	}

	return status;
}

MX_EXPORT mx_status_type
mx_put_field_array_ascii( MX_RECORD *server_record,
			char *remote_record_field_name,
			MX_RECORD_FIELD *local_field,
			void *value_ptr )
{
	const char fname[] = "mx_put_field_array_ascii()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	mx_status_type ( *token_constructor )
		(void *, char *, size_t, MX_RECORD *, MX_RECORD_FIELD *);
	long datatype, num_dimensions, *dimension;

	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header;
	char *buffer;
	char *message, *ptr;
	mx_uint32_type header_length, message_length, message_type, status_code;
	mx_status_type status;

	status = mx_ascii_field_get_parameters( server_record, local_field,
			&datatype, &num_dimensions, &dimension,
			&server, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_get_token_constructor( datatype, &token_constructor );

	if ( status.code != MXE_SUCCESS )
		return status;

	/************ Construct a 'put array ascii' command. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->header[0]);
	buffer = &(aligned_buffer->buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_MESSAGE_TYPE]
				= htonl( MX_NETMSG_PUT_ARRAY_ASCII );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;

	sprintf( message, "%*s",
		-MXU_RECORD_FIELD_NAME_LENGTH, remote_record_field_name );

	ptr = message + MXU_RECORD_FIELD_NAME_LENGTH;

	/* Construct the string description of the array. */

	if ( (num_dimensions == 0)
		|| ((datatype == MXFT_STRING) && (num_dimensions == 1)) ) {

		status = (*token_constructor) ( value_ptr, ptr,
			MX_NETWORK_MAXIMUM_MESSAGE_SIZE - ( ptr - buffer ),
			NULL, local_field );
	} else {
		status = mx_create_array_description( value_ptr, 
			local_field->num_dimensions - 1, ptr,
			MX_NETWORK_MAXIMUM_MESSAGE_SIZE - ( ptr - buffer ),
			NULL, local_field, token_constructor );
	}

	if ( status.code != MXE_SUCCESS )
		return status;

	/* The extra 1 is for the '\0' byte at the end of the string. */

	message_length = 1 + strlen( message );

	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( message_length );

	message_length += MX_NETWORK_HEADER_LENGTH_VALUE;

	MX_DEBUG( 2,("%s: message = '%s'", fname, message));

#if NETWORK_DEBUG
	mx_network_display_message_buffer( buffer );
#endif

	/*************** Send the message. **************/

	status = mx_network_send_message( server_record, buffer );

	if ( status.code != MXE_SUCCESS )
		return status;

	/************** Wait for the response. ************/

	status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

	if ( status.code != MXE_SUCCESS )
		return status;

#if NETWORK_DEBUG
	mx_network_display_message_buffer( buffer );
#endif

        header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
        message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
        message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
        status_code = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

        if ( message_type != mx_server_response(MX_NETMSG_PUT_ARRAY_ASCII) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {
			return mx_error( (long)status_code, fname,
				"%s", message );
		}

                return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_PUT_ARRAY_ASCII_RESPONSE.  "
"Instead it was of type = %ld", message_type );
        }

        message = buffer + header_length;

        /* If the remote command failed, the message field will include
         * the text of the error message rather than the array data
         * we wanted.
         */

        if ( status_code != MXE_SUCCESS ) {
                return mx_error( (long)status_code, fname, message );
        } else {
		return MX_SUCCESSFUL_RESULT;
	}
}

MX_EXPORT mx_status_type
mx_get_field_type( MX_RECORD *server_record, 
			char *remote_record_field_name,
			long max_dimensions,
			long *datatype,
			long *num_dimensions,
			long *dimension_array )
{
	const char fname[] = "mx_get_field_type()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header;
	char *buffer, *message;
	mx_uint32_type *message_uint32_array;
	mx_uint32_type header_length, message_length;
	mx_uint32_type message_type, status_code;
	mx_uint32_type i, expected_message_length;
	mx_uint32_type header_length_in_32bit_words;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked, remote_record_field_name = '%s'",
			fname, remote_record_field_name));

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_record argument passed was NULL." );
	}

	server = (MX_NETWORK_SERVER *) server_record->record_class_struct;

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	/************ Send the 'get field type' message. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->header[0]);
	buffer = &(aligned_buffer->buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_MESSAGE_TYPE] = htonl( MX_NETMSG_GET_FIELD_TYPE );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;

	mx_strncpy( message, remote_record_field_name,
					MXU_RECORD_FIELD_NAME_LENGTH );

	message_length = strlen( message ) + 1;
	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( message_length );

	status = mx_network_send_message( server_record, buffer );

	if ( status.code != MXE_SUCCESS )
		return status;

	/************** Wait for the response. **************/

	status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

	if ( status.code != MXE_SUCCESS )
		return status;

	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	if ( message_type != mx_server_response( MX_NETMSG_GET_FIELD_TYPE ) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {
			return mx_error( (long)status_code, fname,
				"%s", message );
		}

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_GET_FIELD_TYPE_RESPONSE.  "
"Instead it was of type = %ld", message_type );
	}

#if 0
	mx_network_display_message_buffer( buffer );
#endif

	/* If the remote command failed, the message field will include
	 * the text of the error message rather than the field type
	 * data we requested.
	 */

	if ( status_code != MXE_SUCCESS ) {
		return mx_error( (long)status_code, fname, message );
	}

	if ( message_length < (2 * sizeof(long)) ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Incomplete message received.  Message length = %ld",
			message_length );
	}

	/***** Get the type information out of what we were sent. *****/

	header_length_in_32bit_words = header_length / sizeof(mx_uint32_type);

	message_uint32_array = header + header_length_in_32bit_words;

	*datatype = (long) ntohl( (unsigned long) message_uint32_array[0] );
	*num_dimensions
		= (long) ntohl( (unsigned long) message_uint32_array[1] );

	expected_message_length = sizeof(long) * ( (*num_dimensions) + 2 );

	if ( message_length < expected_message_length ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Incomplete dimension array received.  %ld dimension values were received, "
"but %ld dimension values were expected.",
			message_length - 2, *num_dimensions );
	}
	if ( *num_dimensions < max_dimensions ) {
		max_dimensions = *num_dimensions;
	}
	if ( max_dimensions <= 0 ) {
		dimension_array[0] = (long)
			ntohl( (unsigned long) message_uint32_array[2] );
	}
	for ( i = 0; i < max_dimensions; i++ ) {
		dimension_array[i] = (long)
			ntohl( (unsigned long) message_uint32_array[i+2] );
	}
	return MX_SUCCESSFUL_RESULT;
}

