/*
 * Name:    mx_scaler.c
 *
 * Purpose: MX function library of generic scaler operations.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_measurement.h"
#include "mx_timer.h"
#include "mx_scaler.h"

/*=======================================================================*/

/* This function is a utility function to consolidate all of the pointer
 * mangling that often has to happen at the beginning of an mx_scaler_...
 * function.  The parameter 'calling_fname' is passed so that error messages
 * will appear with the name of the calling function.
 */

MX_EXPORT mx_status_type
mx_scaler_get_pointers( MX_RECORD *scaler_record,
			MX_SCALER **scaler,
			MX_SCALER_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_scaler_get_pointers()";

	if ( scaler_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The scaler_record pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( scaler_record->mx_class != MXC_SCALER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"The record '%s' passed by '%s' is not a scaler.",
			scaler_record->name, calling_fname );
	}

	if ( scaler != (MX_SCALER **) NULL ) {
		*scaler = (MX_SCALER *) (scaler_record->record_class_struct);

		if ( *scaler == (MX_SCALER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SCALER pointer for record '%s' passed by '%s' is NULL",
				scaler_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_SCALER_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_SCALER_FUNCTION_LIST *)
				(scaler_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_SCALER_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_SCALER_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				scaler_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_scaler_clear( MX_RECORD *scaler_record )
{
	const char fname[] = "mx_scaler_clear()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *clear_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	clear_fn = function_list->clear;

	if ( clear_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"clear function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	status = (*clear_fn)( scaler );

	/* The "overflow_set" field should _not_ be set to FALSE here,
	 * since for some scaler types, mx_scaler_read_and_clear() is
	 * implemented as the equivalent of a call to mx_scaler_read()
	 * by the equivalent of a call to mx_scaler_clear().
	 */

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_overflow_set( MX_RECORD *scaler_record, int *overflow_set )
{
	const char fname[] = "mx_scaler_overflow_set()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *overflow_set_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	overflow_set_fn = function_list->overflow_set;

	if ( overflow_set_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"overflow_set function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	status = (*overflow_set_fn)( scaler );

	*overflow_set = scaler->overflow_set;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_read( MX_RECORD *scaler_record, long *value )
{
	const char fname[] = "mx_scaler_read()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *read_fn ) ( MX_SCALER * );
	int overflow_set;
	mx_status_type status, status2;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	read_fn = function_list->read;

	if ( read_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	*value = 0L;

	status = (*read_fn)( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Did the scaler overflow? 
	 *
	 * Note there is a race condition here if the scaler is
	 * currently gated on.
	 */

	if ( scaler->automatic_overflow_check ) {

		status2 = mx_scaler_overflow_set(scaler_record, &overflow_set);

		switch ( status2.code ) {
		case MXE_SUCCESS:    /* do nothing */
			break;
		case MXE_UNSUPPORTED:
			status2 = MX_SUCCESSFUL_RESULT;
			overflow_set = FALSE;
			break;
		default:
			return status2;
		}
	}

	*value = scaler->value;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_read_and_clear( MX_RECORD *scaler_record, long *value )
{
	const char fname[] = "mx_scaler_read_and_clear()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *read_and_clear_fn ) ( MX_SCALER * );
	int overflow_set;
	mx_status_type status, status2;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	read_and_clear_fn = function_list->read_and_clear;

	if ( read_and_clear_fn == NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read_and_clear function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	*value = 0L;

	status = (*read_and_clear_fn)( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Did the scaler overflow? 
	 *
	 * Note there is a race condition here if the scaler is
	 * currently gated on.
	 */

	if ( scaler->automatic_overflow_check ) {

		status2 = mx_scaler_overflow_set(scaler_record, &overflow_set);

		switch ( status2.code ) {
		case MXE_SUCCESS:    /* do nothing */
			break;
		case MXE_UNSUPPORTED:
			status2 = MX_SUCCESSFUL_RESULT;
			overflow_set = FALSE;
			break;
		default:
			return status2;
		}
	}

	*value = scaler->value;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_is_busy( MX_RECORD *scaler_record, int *busy )
{
	const char fname[] = "mx_scaler_is_busy()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *busy_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	busy_fn = function_list->scaler_is_busy;

	if ( busy_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"busy function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	status = (*busy_fn)( scaler );

	*busy = scaler->busy;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_start( MX_RECORD *scaler_record, long preset_count )
{
	const char fname[] = "mx_scaler_start()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *start_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	start_fn = function_list->start;

	if ( start_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	scaler->value = preset_count;

	status = (*start_fn)( scaler );

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_stop( MX_RECORD *scaler_record, long *present_value )
{
	const char fname[] = "mx_scaler_stop()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *stop_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	stop_fn = function_list->stop;

	if ( stop_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"stop function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	status = (*stop_fn)( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	*present_value = scaler->value;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_get_mode( MX_RECORD *scaler_record, int *mode )
{
	const char fname[] = "mx_scaler_get_mode()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *get_mode_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_mode_fn = function_list->get_mode;

	if ( get_mode_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_mode function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	status = (*get_mode_fn)( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	*mode = scaler->mode;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_set_mode( MX_RECORD *scaler_record, int mode )
{
	const char fname[] = "mx_scaler_set_mode()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *set_mode_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	set_mode_fn = function_list->set_mode;

	if ( set_mode_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_mode function ptr for scaler record '%s' is NULL.",
			scaler_record->name );
	}

	scaler->mode = mode;

	status = (*set_mode_fn)( scaler );

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_set_modes_of_associated_counters( MX_RECORD *scaler_record )
{
	const char fname[] = "mx_scaler_set_modes_of_associated_counters()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *set_modes_of_associated_counters ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	set_modes_of_associated_counters =
			function_list->set_modes_of_associated_counters;

	/* Do nothing if the driver does not define this function. */

	if ( set_modes_of_associated_counters == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	status = (*set_modes_of_associated_counters)( scaler );

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_get_dark_current( MX_RECORD *scaler_record, double *dark_current )
{
	const char fname[] = "mx_scaler_get_dark_current()";

	MX_SCALER *scaler;
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	*dark_current = scaler->dark_current;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_scaler_set_dark_current( MX_RECORD *scaler_record, double dark_current )
{
	const char fname[] = "mx_scaler_set_dark_current()";

	MX_SCALER *scaler;
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	scaler->dark_current = dark_current;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_scaler_read_dark_current_subtracted_value( MX_RECORD *scaler_record,
						MX_RECORD *timer_record,
						long *value )
{
	const char fname[] = "mx_scaler_read_dark_current_subtracted_value()";

	MX_SCALER *scaler;
	MX_TIMER *timer;
	long offset;
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The timer_record pointer passed was NULL." );
	}
	if ( timer_record->mx_class != MXC_TIMER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The record '%s' passed as the timer_record argument is not a timer.",
			timer_record->name );
	}

	timer = (MX_TIMER *) timer_record->record_class_struct;

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_TIMER pointer for timer record '%s' is NULL.",
			timer_record->name );
	}

	/* This function is only allowed if the last measurement was a 
	 * preset time measurement.  If it was, then the current timer
	 * mode should be MXCM_PRESET_MODE.
	 */

	if ( timer->mode != MXCM_PRESET_MODE ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"This function is only allowed if the timer '%s' is configured "
	"as an preset counter.", timer_record->name );
	}

	/*
	 * In addition, this function is only allowed if a measurement time
	 * greater than zero has been recorded since the last preset time
	 * measurement for this timer.
	 */

	if ( timer->last_measurement_time <= 0.0 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"The timer '%s' has not made a measurement for greater than zero "
	"seconds since it was configured for preset time mode.",
			timer_record->name );
	}

	/* Read the raw measurement. */

	status = mx_scaler_read( scaler_record, value );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Compute and subtract the dark current offset for this measurement. */

	offset = mx_round( scaler->dark_current
				* timer->last_measurement_time );

	*value = *value - offset;

	return MX_SUCCESSFUL_RESULT;
}

