/*
 * Name:    v_aps_topup.c
 *
 * Purpose: Support for the Advanced Photon Source top-up interlock signals.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2000-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include <stdlib.h>

/* The following three include files are from EPICS. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

#include "mx_record.h"
#include "mx_driver.h"
#include "mx_epics.h"
#include "mx_variable.h"
#include "v_aps_topup.h"

MX_RECORD_FUNCTION_LIST mxv_aps_topup_record_function_list = {
	mxv_aps_topup_initialize_type,
	mxv_aps_topup_create_record_structures,
	mxv_aps_topup_finish_record_initialization,
	mxv_aps_topup_delete_record,
	NULL,
	mx_receive_variable,
	mxv_aps_topup_dummy_function,
	mx_receive_variable,
	mxv_aps_topup_dummy_function
};

MX_VARIABLE_FUNCTION_LIST mxv_aps_topup_variable_function_list = {
	mxv_aps_topup_send_variable,
	mxv_aps_topup_receive_variable
};

/*---*/

MX_RECORD_FIELD_DEFAULTS mxv_aps_topup_interlock_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_APS_TOPUP_STANDARD_FIELDS,
	MX_VARIABLE_STANDARD_FIELDS,
	MX_INT_VARIABLE_STANDARD_FIELDS
};

long mxv_aps_topup_interlock_num_record_fields
	= sizeof( mxv_aps_topup_interlock_record_field_defaults )
	/ sizeof( mxv_aps_topup_interlock_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxv_aps_topup_interlock_field_def_ptr
		= &mxv_aps_topup_interlock_record_field_defaults[0];

/*---*/

MX_RECORD_FIELD_DEFAULTS mxv_aps_topup_time_to_inject_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_APS_TOPUP_STANDARD_FIELDS,
	MX_VARIABLE_STANDARD_FIELDS,
	MX_DOUBLE_VARIABLE_STANDARD_FIELDS
};

long mxv_aps_topup_time_to_inject_num_record_fields
	= sizeof( mxv_aps_topup_time_to_inject_record_field_defaults )
	/ sizeof( mxv_aps_topup_time_to_inject_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxv_aps_topup_time_to_inject_field_def_ptr
		= &mxv_aps_topup_time_to_inject_record_field_defaults[0];

/********************************************************************/

MX_EXPORT mx_status_type
mxv_aps_topup_initialize_type( long record_type )
{
	return mx_variable_initialize_type( record_type );
}

MX_EXPORT mx_status_type
mxv_aps_topup_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxv_aps_topup_create_record_structures()";

	MX_VARIABLE *variable_struct;
	MX_APS_TOPUP *aps_topup;

	if ( record->mx_type == MXV_CAL_APS_TOPUP_INTERLOCK ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
	"The topup interlock variable type is not yet implemented.  "
	"Use the topup time to inject variable type instead." );
	}

	/* Allocate memory for the necessary structures. */

	variable_struct = (MX_VARIABLE *) malloc( sizeof(MX_VARIABLE) );

	if ( variable_struct == (MX_VARIABLE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_VARIABLE structure." );
	}

	aps_topup = (MX_APS_TOPUP *)
				malloc( sizeof(MX_APS_TOPUP) );

	if ( aps_topup == (MX_APS_TOPUP *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_APS_TOPUP structure." );
	}

	/* Now set up the necessary pointers. */

	variable_struct->record = record;

	record->record_superclass_struct = variable_struct;
	record->record_class_struct = NULL;
	record->record_type_struct = aps_topup;

	record->superclass_specific_function_list =
				&mxv_aps_topup_variable_function_list;
	record->class_specific_function_list = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_aps_topup_finish_record_initialization( MX_RECORD *record )
{
	/* Turn off automatic printing of EZCA error messages, since we
	 * want to intercept them ourselves.
	 */

	ezcaAutoErrorMessageOff();

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_aps_topup_delete_record( MX_RECORD *record )
{
	const char fname[] = "mxv_aps_topup_delete_record()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxv_aps_topup_dummy_function( MX_RECORD *record )
{
	/* This function serves as a placeholder for functions that are
	 * not needed by this variable class.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_aps_topup_send_variable( MX_VARIABLE *variable )
{
	const char fname[] = "mxv_aps_topup_send_variable()";

	return mx_error( MXE_PERMISSION_DENIED, fname,
	"The APS top-up status variables are read-only." );
}

MX_EXPORT mx_status_type
mxv_aps_topup_receive_variable( MX_VARIABLE *variable )
{
	const char fname[] = "mxv_aps_topup_receive_variable()";

	MX_APS_TOPUP *aps_topup;
	MX_RECORD_FIELD *value_field;
	void *value_ptr;
	char mx_error_message[150];
	char *ezca_error_message;
	char pvname[80];
	int ezca_status;
	double topup_time_to_inject;
	double *double_ptr;
	mx_status_type status;

	aps_topup = (MX_APS_TOPUP *) variable->record->record_type_struct;

	status = mx_find_record_field(variable->record, "value", &value_field);

	if ( status.code != MXE_SUCCESS )
		return status;

	value_ptr = mx_get_field_value_pointer( value_field );

	double_ptr = (double *) value_ptr;

	strcpy( pvname, "Mt:TopUpTime2Inject" );

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &topup_time_to_inject );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );

		sprintf( mx_error_message,
		"Error getting current value of EPICS process variable '%s'.  "
		"EZCA status = %d (%s)",
			pvname, ezca_status,
			ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	*double_ptr = topup_time_to_inject + aps_topup->extra_delay_time;

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

