/*
 * Name:    v_xia_dxp.c
 *
 * Purpose: Support for variable values derived from the XIA DXP
 *          multichannel analyzer.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_record.h"
#include "mx_types.h"
#include "mx_variable.h"
#include "mx_mca.h"
#include "d_xia_dxp.h"
#include "v_xia_dxp.h"

MX_RECORD_FUNCTION_LIST mxv_xia_dxp_record_function_list = {
	mxv_xia_dxp_initialize_type,
	mxv_xia_dxp_create_record_structures,
	mxv_xia_dxp_finish_record_initialization,
	mxv_xia_dxp_delete_record,
	NULL,
	mxv_xia_dxp_dummy_function,
	mxv_xia_dxp_dummy_function,
	mxv_xia_dxp_dummy_function,
	mxv_xia_dxp_dummy_function
};

MX_VARIABLE_FUNCTION_LIST mxv_xia_dxp_variable_function_list = {
	mxv_xia_dxp_send_variable,
	mxv_xia_dxp_receive_variable
};

MX_RECORD_FIELD_DEFAULTS mxv_xia_dxp_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_XIA_DXP_VARIABLE_STANDARD_FIELDS,
	MX_VARIABLE_STANDARD_FIELDS,
	MX_DOUBLE_VARIABLE_STANDARD_FIELDS
};

long mxv_xia_dxp_num_record_fields
	= sizeof( mxv_xia_dxp_record_field_defaults )
	/ sizeof( mxv_xia_dxp_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxv_xia_dxp_rfield_def_ptr
		= &mxv_xia_dxp_record_field_defaults[0];

#define MXV_XIA_DXP_DEBUG	TRUE

/********************************************************************/

MX_EXPORT mx_status_type
mxv_xia_dxp_initialize_type( long record_type )
{
	return mx_variable_initialize_type( record_type );
}

MX_EXPORT mx_status_type
mxv_xia_dxp_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxv_xia_dxp_create_record_structures()";

	MX_VARIABLE *variable_struct;
	MX_XIA_DXP_VARIABLE *xia_dxp_variable;

	/* Allocate memory for the necessary structures. */

	variable_struct = (MX_VARIABLE *) malloc( sizeof(MX_VARIABLE) );

	if ( variable_struct == (MX_VARIABLE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_VARIABLE structure." );
	}

	xia_dxp_variable = (MX_XIA_DXP_VARIABLE *)
				malloc( sizeof(MX_XIA_DXP_VARIABLE) );

	if ( xia_dxp_variable == (MX_XIA_DXP_VARIABLE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_XIA_DXP_VARIABLE structure." );
	}

	/* Now set up the necessary pointers. */

	variable_struct->record = record;

	record->record_superclass_struct = variable_struct;
	record->record_class_struct = NULL;
	record->record_type_struct = xia_dxp_variable;

	record->superclass_specific_function_list =
				&mxv_xia_dxp_variable_function_list;
	record->class_specific_function_list = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_xia_dxp_finish_record_initialization( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_xia_dxp_delete_record( MX_RECORD *record )
{
	const char fname[] = "mxv_xia_dxp_delete_record()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxv_xia_dxp_dummy_function( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_xia_dxp_send_variable( MX_VARIABLE *variable )
{
	const char fname[] = "mxv_xia_dxp_send_variable()";

	return mx_error( MXE_PERMISSION_DENIED, fname,
	"The XIA DXP MCA variables are read-only." );
}

MX_EXPORT mx_status_type
mxv_xia_dxp_receive_variable( MX_VARIABLE *variable )
{
	const char fname[] = "mxv_xia_dxp_receive_variable()";

	MX_RECORD *mca_record;
	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;

	MX_XIA_DXP_VARIABLE *xia_dxp_variable;
	MX_RECORD_FIELD *value_field;
	void *value_ptr;
	double *double_ptr;

	mx_uint32_type parameter_value;
	double corrected_value, multiplier;
	unsigned long mca_value;
	mx_status_type mx_status;

	if ( variable == (MX_VARIABLE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_VARIABLE pointer passed was NULL." );
	}
	if ( variable->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_RECORD pointer for the MX_VARIABLE pointer passed was NULL." );
	}

	xia_dxp_variable = (MX_XIA_DXP_VARIABLE *)
					variable->record->record_type_struct;

	if ( xia_dxp_variable == (MX_XIA_DXP_VARIABLE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_DXP_VARIABLE pointer for variable '%s' is NULL.",
			variable->record->name );
	}

	/* Construct a pointer to the return value. */

	mx_status = mx_find_record_field( variable->record,
						"value", &value_field );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	value_ptr = mx_get_field_value_pointer( value_field );

	double_ptr = (double *) value_ptr;

	/* Construct pointers to the XIA DXP MCA data structures. */

	mca_record = xia_dxp_variable->mca_record;

	if ( mca_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mca_record pointer for variable '%s' is NULL.",
			variable->record->name );
	}

	mca = (MX_MCA *) mca_record->record_class_struct;

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCA pointer for MCA record '%s' is NULL.",
			mca_record->name );
	}

	xia_dxp_mca = (MX_XIA_DXP_MCA *) mca_record->record_type_struct;

	if ( xia_dxp_mca == (MX_XIA_DXP_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_DXP_MCA pointer for MCA record '%s' is NULL.",
			mca_record->name );
	}

	/**** Handle the different value types. ****/

	switch( xia_dxp_variable->value_type ) {

	case MXT_XIA_DXP_VAR_PARAMETER:
		mx_status = (xia_dxp_mca->read_parameter)( mca,
					xia_dxp_variable->value_parameters,
					&parameter_value,
					MXV_XIA_DXP_DEBUG );

		*double_ptr = (double) parameter_value;
		break;

	case MXT_XIA_DXP_VAR_ROI_INTEGRAL:
		mx_status = mx_mca_get_roi_integral(
				xia_dxp_variable->mca_record, &mca_value );

		*double_ptr = (double) mca_value;
		break;

	case MXT_XIA_DXP_VAR_CORRECTED_ROI_INTEGRAL:
		mx_status = mx_mca_get_roi_integral(
				xia_dxp_variable->mca_record, &mca_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		corrected_value = (double) mca_value;

		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		multiplier = mx_divide_safely( xia_dxp_mca->input_count_rate,
					xia_dxp_mca->output_count_rate );

		corrected_value = mx_multiply_safely( corrected_value,
							multiplier );

		multiplier = mx_divide_safely( mca->real_time, mca->live_time );

		corrected_value = mx_multiply_safely( corrected_value,
							multiplier );

		*double_ptr = corrected_value;
		break;

	case MXT_XIA_DXP_VAR_REALTIME:
		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*double_ptr = mca->real_time;
		break;

	case MXT_XIA_DXP_VAR_LIVETIME:
		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*double_ptr = mca->live_time;
		break;

	case MXT_XIA_DXP_VAR_INPUT_COUNT_RATE:
		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*double_ptr = xia_dxp_mca->input_count_rate;
		break;

	case MXT_XIA_DXP_VAR_OUTPUT_COUNT_RATE:
		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*double_ptr = xia_dxp_mca->output_count_rate;
		break;

	case MXT_XIA_DXP_VAR_NUM_FAST_PEAKS:
		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*double_ptr = (double) xia_dxp_mca->num_fast_peaks;
		break;

	case MXT_XIA_DXP_VAR_NUM_EVENTS:
		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*double_ptr = (double) xia_dxp_mca->num_events;
		break;

	case MXT_XIA_DXP_VAR_NUM_UNDERFLOWS:
		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*double_ptr = (double) xia_dxp_mca->num_underflows;
		break;

	case MXT_XIA_DXP_VAR_NUM_OVERFLOWS:
		mx_status = (xia_dxp_mca->read_statistics)( mca,
							MXV_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*double_ptr = (double) xia_dxp_mca->num_overflows;
		break;

	default:
		return mx_error( MXE_UNSUPPORTED, fname,
		"The value type %d for XIA DXP variable '%s' is unsupported.",
			xia_dxp_variable->value_type, variable->record->name );
		break;
	}

	return mx_status;
}

