/*
 * Name:    mmeasure.c
 *
 * Purpose: Measure scaler dark currents or amplifier offsets.
 *
 * Author:  William Lavender
 *
 * Currently this function expects that the scalers are gated on and off
 * by the timer called "timer1".  Probably, there should be something 
 * defined in the MX_SCALER structure that tells you which timer controls
 * which scaler, but currently there is not.
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "motor.h"
#include "mdialog.h"
#include "mx_autoscale.h"
#include "d_auto_scaler.h"

#define FREE_SCALER_ARRAYS \
	do { \
		if ( scaler_array != NULL ) { \
			free( scaler_array ); \
			scaler_array = NULL; \
		} \
		if ( scaler_sum_array != NULL ) { \
			free( scaler_sum_array ); \
			scaler_sum_array = NULL; \
		} \
	} while(0)

static int motor_measure_dark_currents( int argc, char *argv[] );
static int motor_measure_autoscale( int argc, char *argv[] );
static int motor_measure_scaler_offsets( long, MX_RECORD **,
				long *, MX_RECORD *, double, long );

int
motor_measure_fn( int argc, char *argv[] )
{
	static char usage[] =
"Usage:  measure dark_currents [measurement time] [# of measurements]\n"
"        measure autoscale [] []\n";

	size_t length;
	int status;

	if ( argc <= 2 ) {
		fprintf( output, "%s\n", usage );
		return FAILURE;
	}

	length = strlen( argv[2] );

	if ( length <= 0 )
		length = 1;

	if ( (strncmp( argv[2], "dark_currents", length ) == 0)
	  || (strncmp( argv[2], "pedestals", length ) == 0) )
	{
		status = motor_measure_dark_currents( argc, argv );
	} else
	if ( strncmp( argv[2], "autoscale", length ) == 0) {

		status = motor_measure_autoscale( argc, argv );
	} else {
		fprintf( output, "%s", usage );
		status = FAILURE;
	}

	return status;
}

static int
motor_measure_dark_currents( int argc, char *argv[] )
{
	const char cname[] = "motor_measure_dark_currents()";

	MX_RECORD *timer_record;
	MX_RECORD *current_record;
	MX_RECORD **scaler_array;
	MX_SCALER *scaler;
	long *scaler_sum_array;
	long i, num_scalers;
	double measurement_time;
	long num_measurements;
	double new_dark_current;
	int status;
	mx_status_type mx_status;

	scaler_array = NULL;
	scaler_sum_array = NULL;

	/* Verify that the "timer1" record exists, since we can't make
	 * a measurement without it.
	 */

	timer_record = mx_get_record( motor_record_list, "timer1" );

	if ( timer_record == (MX_RECORD *) NULL ) {
		fprintf(output,"The timer record 'timer1' does not exist!\n");
		fprintf(output,"We cannot make a measurement without it.\n");
		return FAILURE;
	}

	/* Get a list of all the scaler records. */

	/* First, how many scalers are there? */

	num_scalers = 0L;

	current_record = motor_record_list->next_record;

	while ( current_record != motor_record_list ) {

		if ( current_record->mx_class == MXC_SCALER ) {

			scaler = (MX_SCALER *)
				current_record->record_class_struct;

			if ( scaler == (MX_SCALER *) NULL ) {
				fprintf( output,
			"The MX_SCALER pointer for scaler '%s' is NULL.",
					current_record->name );

				return FAILURE;
			}
			if ( scaler->subtract_dark_current != 0 )
				num_scalers++;
		}

		current_record = current_record->next_record;
	}

	/* Allocate an array to store the list of scalers. */

	scaler_array = (MX_RECORD **) malloc(num_scalers * sizeof(MX_RECORD *));

	if ( scaler_array == (MX_RECORD **) NULL ) {
		fprintf( output,
		"Ran out of memory trying to allocate an array of %ld scaler "
		"record pointers.\n", num_scalers );
		return FAILURE;
	}

	scaler_sum_array = (long *) malloc( num_scalers * sizeof(long) );

	if ( scaler_sum_array == (long *) NULL ) {
		fprintf( output,
"Ran out of memory trying to allocate an array of %ld scaler values.\n",
			num_scalers );
		FREE_SCALER_ARRAYS;
		return FAILURE;
	}

	/* Now fill the array with the MX_SCALER pointers.  Skip any scalers
	 * that have dark current subtraction disabled.
	 */

	i = 0;

	current_record = motor_record_list->next_record;

	while ( (current_record != motor_record_list) && (i < num_scalers) ) {

		if ( current_record->mx_class == MXC_SCALER ) {

			scaler = (MX_SCALER *)
				current_record->record_class_struct;

			if ( scaler == (MX_SCALER *) NULL ) {
				fprintf( output,
			"The MX_SCALER pointer for scaler '%s' is NULL.",
					current_record->name );

				return FAILURE;
			}
			if ( scaler->subtract_dark_current != 0 ) {

				scaler_array[i] = current_record;

				MX_DEBUG(-2,(
				"%s: scaler '%s' will be measured.",
					cname, current_record->name ));

				i++;
			}
		}
		current_record = current_record->next_record;
	}

	if ( i < num_scalers ) {
		fprintf( output,
"Warning: The number of scalers changed during startup of this command.\n" );
		fprintf( output, "num_scalers = %ld, i = %ld\n",num_scalers,i);
	}

#if 0
	fprintf( output, "Number of scalers = %ld\n", num_scalers );
#endif

	/* Prompt the user for the measurement parameters. */

	if ( argc > 3 ) {
		measurement_time = atof( argv[3] );
	} else {
		measurement_time = 0.0;

		status = motor_get_double( output,
			"Enter measurement time (seconds) -> ",
			TRUE, 1.0,
			&measurement_time, 0.0, 1.0e38 );

		if ( status != SUCCESS ) {
			FREE_SCALER_ARRAYS;
			return status;
		}
	}

	if ( argc > 4 ) {
		num_measurements = atol( argv[4] );
	} else {
		num_measurements = 0L;

		status = motor_get_long( output,
			"Enter number of measurements to average -> ",
			TRUE, 1,
			&num_measurements, 1L, 1000000L );

		if ( status != SUCCESS ) {
			FREE_SCALER_ARRAYS;
			return status;
		}
	}

	/***** Perform the measurements. *****/

	status = motor_measure_scaler_offsets( num_scalers, scaler_array,
				scaler_sum_array, timer_record,
				measurement_time, num_measurements );

	if ( status != SUCCESS ) {
		FREE_SCALER_ARRAYS;
		return status;
	}

	/* Now save the new dark_current values. */

	for ( i = 0; i < num_scalers; i++ ) {
		new_dark_current = (double) scaler_sum_array[i]
			/ (measurement_time * (double) num_measurements);

		mx_status = mx_scaler_set_dark_current( scaler_array[i],
					new_dark_current );

		if ( mx_status.code != MXE_SUCCESS ) {
			FREE_SCALER_ARRAYS;
			return FAILURE;
		}

		fprintf( output,
			"Dark current for '%s' = %.*g counts per second.\n",
			scaler_array[i]->name,
			scaler_array[i]->precision,
			new_dark_current );
	}

	FREE_SCALER_ARRAYS;

	return SUCCESS;
}

static int
motor_measure_autoscale( int argc, char *argv[] )
{
	static char cname[] = "motor_measure_autoscale()";

	MX_RECORD *autoscale_scaler_record;
	MX_AUTOSCALE_SCALER *autoscale_scaler;
	MX_AUTOSCALE *autoscale;
	char *autoscale_scaler_name;
	double measurement_time;
	long num_measurements, scaler_sum;
	unsigned long i, num_monitor_offsets, old_monitor_offset_index;
	double *monitor_offset_array;
	int status;
	mx_status_type mx_status;

	static char usage[] =
"Usage: measure autoscale 'scaler_name' [measurement_time] [# measurements]\n";

	MX_DEBUG(-2,("%s invoked.", cname));

	/**** Parse the arguments to 'measure autoscale'. ****/

	/* The first argument (required) is the name of the autoscale scaler.*/

	if ( argc <= 3 ) {
		fputs(usage, output);
		return FAILURE;
	}

	autoscale_scaler_name = argv[3];

	/* Does this record exist? */

	autoscale_scaler_record = mx_get_record( motor_record_list,
						autoscale_scaler_name );

	if ( autoscale_scaler_record == (MX_RECORD *) NULL ) {
		fprintf(output, "The record '%s' does not exist.\n",
				autoscale_scaler_name );
		return FAILURE;
	}

	/* Is it the right kind of record? */

	if ( (autoscale_scaler_record->mx_superclass != MXR_DEVICE)
	  || (autoscale_scaler_record->mx_class != MXC_SCALER)
	  || (autoscale_scaler_record->mx_type != MXT_SCL_AUTOSCALE) )
	{
		fprintf(output,
		"The record '%s' is not an autoscaling scaler record.\n",
			autoscale_scaler_record->name );
		return FAILURE;
	}

	/* Get the various records that the autoscale scaler depends on. */

	autoscale_scaler = (MX_AUTOSCALE_SCALER *)
				autoscale_scaler_record->record_type_struct;

	if ( autoscale_scaler == (MX_AUTOSCALE_SCALER *) NULL ) {
		fprintf(output,
		"The MX_AUTOSCALE_SCALER pointer for record '%s' is NULL.\n",
			autoscale_scaler_record->name);
		return FAILURE;
	}

	if ( autoscale_scaler->autoscale_record == (MX_RECORD *) NULL ) {
		fprintf(output,
	"The autoscale_record pointer for autoscale scaler '%s' is NULL.\n",
			autoscale_scaler_record->name );
		return FAILURE;
	}

	autoscale = (MX_AUTOSCALE *)
		autoscale_scaler->autoscale_record->record_class_struct;

	if ( autoscale == (MX_AUTOSCALE *) NULL ) {
		fprintf(output,
		"The MX_AUTOSCALE pointer for autoscale record '%s' is NULL.\n",
			autoscale_scaler->autoscale_record->name );
		return FAILURE;
	}

	MX_DEBUG(-2,("%s: autoscale record = '%s'",
		cname, autoscale->record->name ));
	MX_DEBUG(-2,("%s: monitor record   = '%s'",
		cname, autoscale->monitor_record->name));
	MX_DEBUG(-2,("%s: control record   = '%s'",
		cname, autoscale->control_record->name));
	MX_DEBUG(-2,("%s: timer record     = '%s'",
		cname, autoscale->timer_record->name));

	/* Prompt the user for the measurement parameters. */

	if ( argc > 4 ) {
		measurement_time = atof( argv[4] );
	} else {
		measurement_time = 0.0;

		status = motor_get_double( output,
			"Enter measurement time (seconds) -> ",
			TRUE, 1.0,
			&measurement_time, 0.0, 1.0e38 );

		if ( status != SUCCESS )
			return status;
	}

	if ( argc > 5 ) {
		num_measurements = atol( argv[5] );
	} else {
		num_measurements = 0L;

		status = motor_get_long( output,
			"Enter number of measurements to average -> ",
			TRUE, 1,
			&num_measurements, 1L, 1000000L );

		if ( status != SUCCESS )
			return status;
	}

	MX_DEBUG(-2,("%s: measurement_time = %g",
		cname, measurement_time));
	MX_DEBUG(-2,("%s: num_measurements = %ld",
		cname, num_measurements));

	/* Find out how many monitor offsets there are. */

	mx_status = mx_autoscale_get_offset_array( autoscale->record,
						&num_monitor_offsets,
						&monitor_offset_array );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	/* Save the value of which monitor offset is currently being used. */

	mx_status = mx_autoscale_get_offset_index( autoscale->record,
						&old_monitor_offset_index );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	/* Measure the dark currents for all the possible monitor 
	 * offset indices.
	 */

	for ( i = 0; i < num_monitor_offsets; i++ ) {

		mx_status = mx_autoscale_set_offset_index(autoscale->record, i);

		if ( mx_status.code != MXE_SUCCESS )
			return FAILURE;

		status = motor_measure_scaler_offsets( 1L,
					&(autoscale->monitor_record),
					&scaler_sum, autoscale->timer_record,
					measurement_time, num_measurements );

		monitor_offset_array[i] = (double) scaler_sum
			/ (measurement_time * (double) num_measurements);

		MX_DEBUG(-2,(
		"%s: i = %lu, scaler_sum = %ld, dark_current = %g",
			cname, i, scaler_sum, monitor_offset_array[i]));
	}

	/* Set the new values of the monitor offsets. */

	mx_status = mx_autoscale_set_offset_array( autoscale->record,
						num_monitor_offsets,
						monitor_offset_array );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	/* Restore the old monitor offset index. */

	mx_status = mx_autoscale_set_offset_index( autoscale->record,
						old_monitor_offset_index );
	
	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	return SUCCESS;
}

static int
motor_measure_scaler_offsets( long num_scalers,
				MX_RECORD **scaler_record_array,
				long *scaler_sum_array,
				MX_RECORD *timer_record,
				double measurement_time,
				long num_measurements )
{
	double seconds_left;
	long i, n, value;
	int busy;
	mx_status_type mx_status;

	/* First, zero the scaler value array. */

	for ( i = 0; i < num_scalers; i++ ) {
		scaler_sum_array[i] = 0L;
	}

	/* Stop the timer if it is on. */

	mx_status = mx_timer_stop( timer_record, &seconds_left );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	/* Loop through the number of measurements. */

	for ( n = 0; n < num_measurements; n++ ) {

		/* Clear all the scalers. */

		for ( i = 0; i < num_scalers; i++ ) {
			mx_status = mx_scaler_clear( scaler_record_array[i] );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;
		}

		/* Turn the timer on and wait for it to finish. */

		mx_status = mx_timer_start( timer_record, measurement_time );

		if ( mx_status.code != MXE_SUCCESS )
			return FAILURE;

		busy = TRUE;

		while ( busy ) {
			if ( mx_user_requested_interrupt() ) {
				fprintf( output,
				"Dark current measurement was interrupted.\n");
				return FAILURE;
			}

			mx_status = mx_timer_is_busy( timer_record, &busy );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;

			mx_msleep(1);	/* Wait at least 1 millisecond. */
		}

		/* Read out the scalers. */

		for ( i = 0; i < num_scalers; i++ ) {

			mx_status = mx_scaler_read( scaler_record_array[i],
							&value );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;

			scaler_sum_array[i] += value;
		}
	}

	return SUCCESS;
}

