/*
 * Name:    ms_process_record.c
 *
 * Purpose: Functions used to process generic MX events.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_driver.h"
#include "mx_record.h"
#include "mx_motor.h"
#include "mx_socket.h"

#include "ms_server.h"
#include "ms_process_record.h"

mx_status_type
mxsrv_setup_record_process_functions( MX_RECORD *record_list_head )
{
	const char fname[] = "mxsrv_setup_record_process_functions()";

	MX_RECORD *current_record;
	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* The following is done just in case we were called with
	 * a record that isn't really the list head.
	 */

	record_list_head = record_list_head->list_head;

	/* Walk through the record list. */

	current_record = record_list_head->next_record;

	while ( current_record != record_list_head ) {

		record_field_array = current_record->record_field_array;

		for ( i = 0; i < current_record->num_record_fields; i++ ) {

			record_field = &record_field_array[i];

			switch( record_field->label_value ) {
			case MXLV_REC_RESYNCHRONIZE:
				record_field->process_function
					    = mxsrv_record_process_function;
				break;
			default:
				break;
			}
		}
		current_record = current_record->next_record;
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_record_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	const char fname[] = "mxsrv_record_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_RECORD_FUNCTION_LIST *record_function_list;
	mx_status_type ( *resynchronize_fn ) ( MX_RECORD * );
	char operation_string[30];
	long label_value;
	mx_status_type status;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;

	switch ( operation ) {
	case MX_PROCESS_GET:
		strcpy( operation_string, "(MX_PROCESS_GET)" );
		break;
	case MX_PROCESS_PUT:
		strcpy( operation_string, "(MX_PROCESS_PUT)" );
		break;
	default:
		sprintf( operation_string, "(Operation %d)", operation );
		break;
	}

	label_value = record_field->label_value;

	MX_DEBUG(1,("%s: '%s.%s' %s (%ld)",
		fname, record->name, record_field->name,
		operation_string, label_value));

	/*
	 * A label value < 0 tells us that we do not need to do any
	 * further MX record processing. 
	 */

	if ( label_value < 0 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	record_function_list = (MX_RECORD_FUNCTION_LIST *)
					( record->record_function_list );

	status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( label_value ) {
		default:
			MX_DEBUG(-2,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	case MX_PROCESS_PUT:
		switch( label_value ) {
		case MXLV_REC_RESYNCHRONIZE:
			resynchronize_fn = record_function_list->resynchronize;

			if ( resynchronize_fn == NULL ) {

				status = mx_error( MXE_UNSUPPORTED, fname,
			    "Resynchronize is not supported for record '%s'",
					record->name );
			} else {
				status = ( *resynchronize_fn ) ( record );
			}
			break;
		default:
			MX_DEBUG(-2,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %d", operation );
		break;
	}

	return status;
}

